@@ -1,4 +1,19 @@
 const char* mysql_fix_privilege_tables={
+"-- Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.\n "
+"-- \n "
+"-- This program is free software; you can redistribute it and/or modify\n "
+"-- it under the terms of the GNU General Public License as published by\n "
+"-- the Free Software Foundation; version 2 of the License.\n "
+"-- \n "
+"-- This program is distributed in the hope that it will be useful,\n "
+"-- but WITHOUT ANY WARRANTY; without even the implied warranty of\n "
+"-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n "
+"-- GNU General Public License for more details.\n "
+"-- \n "
+"-- You should have received a copy of the GNU General Public License\n "
+"-- along with this program; if not, write to the Free Software\n "
+"-- Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA\n "
+"\n "
 "--\n "
 "-- The system tables of MySQL Server\n "
 "--\n "
@@ -79,12 +94,12 @@ const char* mysql_fix_privilege_tables={
 "ES', 'STRICT_TRANS_TABLES', 'STRICT_ALL_TABLES', 'NO_ZERO_IN_DATE', 'NO_ZERO_DATE', 'INVALID_DATES', 'ERROR_FOR_DIVISION_BY_ZERO', 'TRADITIONAL', 'NO_AUTO_CREATE_USER', 'HIGH_NOT_PRECEDENCE', 'NO_ENGINE_SUBSTITUTION', 'PAD_CHAR_TO_FULL_LENGTH') DEFAULT '' NOT NULL, comment char(64) collate utf8_bin DEFAULT '' NOT NULL, character_set_client char(32) collate utf8_bin, collation_connection char(32) collate utf8_bin, db_collation char(32) collate utf8_bin, body_utf8 longblob, PRIMARY KEY (db,name,type)) engine"
 "=MyISAM character set utf8 comment='Stored Procedures';\n "
 "\n "
-"CREATE TABLE IF NOT EXISTS procs_priv ( Host char(60) binary DEFAULT '' NOT NULL, Db char(64) binary DEFAULT '' NOT NULL, User char(16) binary DEFAULT '' NOT NULL, Routine_name char(64) binary DEFAULT '' NOT NULL, Routine_type enum('FUNCTION','PROCEDURE') NOT NULL, Grantor char(77) DEFAULT '' NOT NULL, Proc_priv set('Execute','Alter Routine','Grant') COLLATE utf8_general_ci DEFAULT '' NOT NULL, Timestamp timestamp(14), PRIMARY KEY (Host,Db,User,Routine_name,Routine_type), KEY Grantor (Grantor) ) engine=MyI"
-"SAM CHARACTER SET utf8 COLLATE utf8_bin   comment='Procedure privileges';\n "
+"CREATE TABLE IF NOT EXISTS procs_priv ( Host char(60) binary DEFAULT '' NOT NULL, Db char(64) binary DEFAULT '' NOT NULL, User char(16) binary DEFAULT '' NOT NULL, Routine_name char(64) COLLATE utf8_general_ci DEFAULT '' NOT NULL, Routine_type enum('FUNCTION','PROCEDURE') NOT NULL, Grantor char(77) DEFAULT '' NOT NULL, Proc_priv set('Execute','Alter Routine','Grant') COLLATE utf8_general_ci DEFAULT '' NOT NULL, Timestamp timestamp(14), PRIMARY KEY (Host,Db,User,Routine_name,Routine_type), KEY Grantor (Gran"
+"tor) ) engine=MyISAM CHARACTER SET utf8 COLLATE utf8_bin   comment='Procedure privileges';\n "
 "\n "
 "-- Create general_log if CSV is enabled.\n "
 "\n "
-"SET @str = IF (@@have_csv = 'YES', 'CREATE TABLE IF NOT EXISTS general_log (event_time TIMESTAMP NOT NULL, user_host MEDIUMTEXT NOT NULL, thread_id INTEGER NOT NULL, server_id INTEGER NOT NULL, command_type VARCHAR(64) NOT NULL, argument MEDIUMTEXT NOT NULL) engine=CSV CHARACTER SET utf8 comment=\"General log\"', 'SET @dummy = 0');\n "
+"SET @str = IF (@@have_csv = 'YES', 'CREATE TABLE IF NOT EXISTS general_log (event_time TIMESTAMP NOT NULL, user_host MEDIUMTEXT NOT NULL, thread_id INTEGER NOT NULL, server_id INTEGER UNSIGNED NOT NULL, command_type VARCHAR(64) NOT NULL, argument MEDIUMTEXT NOT NULL) engine=CSV CHARACTER SET utf8 comment=\"General log\"', 'SET @dummy = 0');\n "
 "\n "
 "PREPARE stmt FROM @str;\n "
 "EXECUTE stmt;\n "
@@ -92,7 +107,7 @@ const char* mysql_fix_privilege_tables={
 "\n "
 "-- Create slow_log if CSV is enabled.\n "
 "\n "
-"SET @str = IF (@@have_csv = 'YES', 'CREATE TABLE IF NOT EXISTS slow_log (start_time TIMESTAMP NOT NULL, user_host MEDIUMTEXT NOT NULL, query_time TIME NOT NULL, lock_time TIME NOT NULL, rows_sent INTEGER NOT NULL, rows_examined INTEGER NOT NULL, db VARCHAR(512) NOT NULL, last_insert_id INTEGER NOT NULL, insert_id INTEGER NOT NULL, server_id INTEGER NOT NULL, sql_text MEDIUMTEXT NOT NULL) engine=CSV CHARACTER SET utf8 comment=\"Slow log\"', 'SET @dummy = 0');\n "
+"SET @str = IF (@@have_csv = 'YES', 'CREATE TABLE IF NOT EXISTS slow_log (start_time TIMESTAMP NOT NULL, user_host MEDIUMTEXT NOT NULL, query_time TIME NOT NULL, lock_time TIME NOT NULL, rows_sent INTEGER NOT NULL, rows_examined INTEGER NOT NULL, db VARCHAR(512) NOT NULL, last_insert_id INTEGER NOT NULL, insert_id INTEGER NOT NULL, server_id INTEGER UNSIGNED NOT NULL, sql_text MEDIUMTEXT NOT NULL) engine=CSV CHARACTER SET utf8 comment=\"Slow log\"', 'SET @dummy = 0');\n "
 "\n "
 "PREPARE stmt FROM @str;\n "
 "EXECUTE stmt;\n "
@@ -101,10 +116,311 @@ const char* mysql_fix_privilege_tables={
 "CREATE TABLE IF NOT EXISTS event ( db char(64) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL default '', name char(64) CHARACTER SET utf8 NOT NULL default '', body longblob NOT NULL, definer char(77) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL default '', execute_at DATETIME default NULL, interval_value int(11) default NULL, interval_field ENUM('YEAR','QUARTER','MONTH','DAY','HOUR','MINUTE','WEEK','SECOND','MICROSECOND','YEAR_MONTH','DAY_HOUR','DAY_MINUTE','DAY_SECOND','HOUR_MINUTE','HOUR_SECOND','MINUTE_S"
 "ECOND','DAY_MICROSECOND','HOUR_MICROSECOND','MINUTE_MICROSECOND','SECOND_MICROSECOND') default NULL, created TIMESTAMP NOT NULL, modified TIMESTAMP NOT NULL, last_executed DATETIME default NULL, starts DATETIME default NULL, ends DATETIME default NULL, status ENUM('ENABLED','DISABLED','SLAVESIDE_DISABLED') NOT NULL default 'ENABLED', on_completion ENUM('DROP','PRESERVE') NOT NULL default 'DROP', sql_mode  set('REAL_AS_FLOAT','PIPES_AS_CONCAT','ANSI_QUOTES','IGNORE_SPACE','NOT_USED','ONLY_FULL_GROUP_BY','NO"
 "_UNSIGNED_SUBTRACTION','NO_DIR_IN_CREATE','POSTGRESQL','ORACLE','MSSQL','DB2','MAXDB','NO_KEY_OPTIONS','NO_TABLE_OPTIONS','NO_FIELD_OPTIONS','MYSQL323','MYSQL40','ANSI','NO_AUTO_VALUE_ON_ZERO','NO_BACKSLASH_ESCAPES','STRICT_TRANS_TABLES','STRICT_ALL_TABLES','NO_ZERO_IN_DATE','NO_ZERO_DATE','INVALID_DATES','ERROR_FOR_DIVISION_BY_ZERO','TRADITIONAL','NO_AUTO_CREATE_USER','HIGH_NOT_PRECEDENCE','NO_ENGINE_SUBSTITUTION','PAD_CHAR_TO_FULL_LENGTH') DEFAULT '' NOT NULL, comment char(64) CHARACTER SET utf8 COLLATE "
-"utf8_bin NOT NULL default '', originator int(10) NOT NULL, time_zone char(64) CHARACTER SET latin1 NOT NULL DEFAULT 'SYSTEM', character_set_client char(32) collate utf8_bin, collation_connection char(32) collate utf8_bin, db_collation char(32) collate utf8_bin, body_utf8 longblob, PRIMARY KEY (db, name) ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COMMENT 'Events';\n "
+"utf8_bin NOT NULL default '', originator INTEGER UNSIGNED NOT NULL, time_zone char(64) CHARACTER SET latin1 NOT NULL DEFAULT 'SYSTEM', character_set_client char(32) collate utf8_bin, collation_connection char(32) collate utf8_bin, db_collation char(32) collate utf8_bin, body_utf8 longblob, PRIMARY KEY (db, name) ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COMMENT 'Events';\n "
 "\n "
 "\n "
-"CREATE TABLE IF NOT EXISTS ndb_binlog_index (Position BIGINT UNSIGNED NOT NULL, File VARCHAR(255) NOT NULL, epoch BIGINT UNSIGNED NOT NULL, inserts INT UNSIGNED NOT NULL, updates INT UNSIGNED NOT NULL, deletes INT UNSIGNED NOT NULL, schemaops INT UNSIGNED NOT NULL, orig_server_id INT UNSIGNED NOT NULL, orig_epoch BIGINT UNSIGNED NOT NULL, gci INT UNSIGNED NOT NULL, PRIMARY KEY(epoch, orig_server_id, orig_epoch)) ENGINE=MYISAM;\n "
+"CREATE TABLE IF NOT EXISTS ndb_binlog_index (Position BIGINT UNSIGNED NOT NULL, File VARCHAR(255) NOT NULL, epoch BIGINT UNSIGNED NOT NULL, inserts INT UNSIGNED NOT NULL, updates INT UNSIGNED NOT NULL, deletes INT UNSIGNED NOT NULL, schemaops INT UNSIGNED NOT NULL, orig_server_id INT UNSIGNED NOT NULL, orig_epoch BIGINT UNSIGNED NOT NULL, gci INT UNSIGNED NOT NULL, next_position BIGINT UNSIGNED NOT NULL, next_file VARCHAR(255) NOT NULL, PRIMARY KEY(epoch, orig_server_id, orig_epoch)) ENGINE=MYISAM;\n "
+"\n "
+"#\n "
+"# SQL commands for creating the tables in MySQL Server which\n "
+"# are used by the NDBINFO storage engine to access system\n "
+"# information and statistics from MySQL Cluster\n "
+"#\n "
+"# Only create objects if NDBINFO is supported\n "
+"SELECT @have_ndbinfo:= COUNT(*) FROM information_schema.engines WHERE engine='NDBINFO' AND support IN ('YES', 'DEFAULT');\n "
+"\n "
+"# Only create objects if version >= 7.1\n "
+"SET @str=IF(@have_ndbinfo,'SELECT @have_ndbinfo:= (@@ndbinfo_version >= (7 << 16) | (1 << 8)) || @ndbinfo_skip_version_check','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# Only create objects if ndbinfo namespace is free\n "
+"SET @str=IF(@have_ndbinfo,'SET @@ndbinfo_show_hidden=TRUE','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'SELECT @have_ndbinfo:= COUNT(*) = 0 FROM information_schema.tables WHERE table_schema = @@ndbinfo_database AND LEFT(table_name, LENGTH(@@ndbinfo_table_prefix)) = @@ndbinfo_table_prefix AND engine != \"ndbinfo\"','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'SET @@ndbinfo_show_hidden=default','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE DATABASE IF NOT EXISTS `ndbinfo`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# Drop any old views in ndbinfo\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.transporters','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.logspaces','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.logbuffers','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.resources','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.counters','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.nodes','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.memoryusage','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP VIEW IF EXISTS ndbinfo.diskpagebuffer','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# Drop any old lookup tables in ndbinfo\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS ndbinfo.blocks','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS ndbinfo.config_params','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$tables\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$tables`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$tables` (`table_id` INT UNSIGNED,`table_name` VARCHAR(512),`comment` VARCHAR(512)) COMMENT=\"metadata for tables available through ndbinfo\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$columns\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$columns`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$columns` (`table_id` INT UNSIGNED,`column_id` INT UNSIGNED,`column_name` VARCHAR(512),`column_type` INT UNSIGNED,`comment` VARCHAR(512)) COMMENT=\"metadata for columns available through ndbinfo \" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$test\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$test`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$test` (`node_id` INT UNSIGNED,`block_number` INT UNSIGNED,`block_instance` INT UNSIGNED,`counter` INT UNSIGNED,`counter2` BIGINT UNSIGNED) COMMENT=\"for testing\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$pools\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$pools`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$pools` (`node_id` INT UNSIGNED,`block_number` INT UNSIGNED,`block_instance` INT UNSIGNED,`pool_name` VARCHAR(512),`used` BIGINT UNSIGNED COMMENT \"currently in use\",`total` BIGINT UNSIGNED COMMENT \"total allocated\",`high` BIGINT UNSIGNED COMMENT \"in use high water mark\",`entry_size` BIGINT UNSIGNED COMMENT \"size in bytes of each object\",`config_param1` INT UNSIGNED COMMENT \"config param 1 affecting pool\",`config_param2` INT UNSIGNED COMMENT \"config para"
+"m 2 affecting pool\",`config_param3` INT UNSIGNED COMMENT \"config param 3 affecting pool\",`config_param4` INT UNSIGNED COMMENT \"config param 4 affecting pool\") COMMENT=\"pool usage\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$transporters\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$transporters`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$transporters` (`node_id` INT UNSIGNED,`remote_node_id` INT UNSIGNED,`connection_status` INT UNSIGNED) COMMENT=\"transporter status\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$logspaces\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$logspaces`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$logspaces` (`node_id` INT UNSIGNED,`log_type` INT UNSIGNED COMMENT \"0 = REDO, 1 = DD-UNDO\",`log_id` INT UNSIGNED,`log_part` INT UNSIGNED,`total` BIGINT UNSIGNED COMMENT \"total allocated\",`used` BIGINT UNSIGNED COMMENT \"currently in use\",`high` BIGINT UNSIGNED COMMENT \"in use high water mark\") COMMENT=\"logspace usage\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$logbuffers\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$logbuffers`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$logbuffers` (`node_id` INT UNSIGNED,`log_type` INT UNSIGNED COMMENT \"0 = REDO, 1 = DD-UNDO\",`log_id` INT UNSIGNED,`log_part` INT UNSIGNED,`total` BIGINT UNSIGNED COMMENT \"total allocated\",`used` BIGINT UNSIGNED COMMENT \"currently in use\",`high` BIGINT UNSIGNED COMMENT \"in use high water mark\") COMMENT=\"logbuffer usage\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$resources\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$resources`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$resources` (`node_id` INT UNSIGNED,`resource_id` INT UNSIGNED,`reserved` INT UNSIGNED COMMENT \"reserved for this resource\",`used` INT UNSIGNED COMMENT \"currently in use\",`max` INT UNSIGNED COMMENT \"max available\",`high` INT UNSIGNED COMMENT \"in use high water mark\") COMMENT=\"resources usage (a.k.a superpool)\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$counters\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$counters`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$counters` (`node_id` INT UNSIGNED,`block_number` INT UNSIGNED,`block_instance` INT UNSIGNED,`counter_id` INT UNSIGNED,`val` BIGINT UNSIGNED COMMENT \"monotonically increasing since process start\") COMMENT=\"monotonic counters\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$nodes\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$nodes`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$nodes` (`node_id` INT UNSIGNED,`uptime` BIGINT UNSIGNED COMMENT \"time in seconds that node has been running\",`status` INT UNSIGNED COMMENT \"starting/started/stopped etc.\",`start_phase` INT UNSIGNED COMMENT \"start phase if node is starting\") COMMENT=\"node status\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.ndb$diskpagebuffer\n "
+"SET @str=IF(@have_ndbinfo,'DROP TABLE IF EXISTS `ndbinfo`.`ndb$diskpagebuffer`','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$diskpagebuffer` (`node_id` INT UNSIGNED,`block_instance` INT UNSIGNED,`pages_written` BIGINT UNSIGNED COMMENT \"Pages written to disk\",`pages_written_lcp` BIGINT UNSIGNED COMMENT \"Pages written by local checkpoint\",`pages_read` BIGINT UNSIGNED COMMENT \"Pages read from disk\",`log_waits` BIGINT UNSIGNED COMMENT \"Page writes waiting for log to be written to disk\",`page_requests_direct_return` BIGINT UNSIGNED COMMENT \"Page in buffer and no requests waiting "
+"for it\",`page_requests_wait_queue` BIGINT UNSIGNED COMMENT \"Page in buffer, but some requests are already waiting for it\",`page_requests_wait_io` BIGINT UNSIGNED COMMENT \"Page not in buffer, waiting to be read from disk\") COMMENT=\"disk page buffer info\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.blocks\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`blocks` (block_number INT UNSIGNED PRIMARY KEY, block_name VARCHAR(512))','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'INSERT INTO `ndbinfo`.`blocks` VALUES (254, \"CMVMI\"), (248, \"DBACC\"), (250, \"DBDICT\"), (246, \"DBDIH\"), (247, \"DBLQH\"), (245, \"DBTC\"), (249, \"DBTUP\"), (253, \"NDBFS\"), (251, \"NDBCNTR\"), (252, \"QMGR\"), (255, \"TRIX\"), (244, \"BACKUP\"), (256, \"DBUTIL\"), (257, \"SUMA\"), (258, \"DBTUX\"), (259, \"TSMAN\"), (260, \"LGMAN\"), (261, \"PGMAN\"), (262, \"RESTORE\"), (263, \"DBINFO\"), (264, \"DBSPJ\")','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.config_params\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`config_params` (param_number INT UNSIGNED PRIMARY KEY, param_name VARCHAR(512))','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"SET @str=IF(@have_ndbinfo,'INSERT INTO `ndbinfo`.`config_params` VALUES (179, \"MaxNoOfSubscriptions\"), (180, \"MaxNoOfSubscribers\"), (181, \"MaxNoOfConcurrentSubOperations\"), (5, \"HostName\"), (3, \"NodeId\"), (101, \"NoOfReplicas\"), (103, \"MaxNoOfAttributes\"), (102, \"MaxNoOfTables\"), (149, \"MaxNoOfOrderedIndexes\"), (150, \"MaxNoOfUniqueHashIndexes\"), (110, \"MaxNoOfConcurrentIndexOperations\"), (105, \"MaxNoOfTriggers\"), (109, \"MaxNoOfFiredTriggers\"), (100, \"MaxNoOfSavedMessages\"), (177, \"LockExecuteThreadToCPU\"), "
+"(178, \"LockMaintThreadsToCPU\"), (176, \"RealtimeScheduler\"), (114, \"LockPagesInMainMemory\"), (123, \"TimeBetweenWatchDogCheck\"), (174, \"SchedulerExecutionTimer\"), (175, \"SchedulerSpinTimer\"), (141, \"TimeBetweenWatchDogCheckInitial\"), (124, \"StopOnError\"), (107, \"MaxNoOfConcurrentOperations\"), (151, \"MaxNoOfLocalOperations\"), (152, \"MaxNoOfLocalScans\"), (153, \"BatchSizePerLocalScan\"), (106, \"MaxNoOfConcurrentTransactions\"), (108, \"MaxNoOfConcurrentScans\"), (111, \"TransactionBufferMemory\"), (113, \"IndexMemory\""
+"), (112, \"DataMemory\"), (154, \"UndoIndexBuffer\"), (155, \"UndoDataBuffer\"), (156, \"RedoBuffer\"), (157, \"LongMessageBuffer\"), (160, \"DiskPageBufferMemory\"), (198, \"SharedGlobalMemory\"), (115, \"StartPartialTimeout\"), (116, \"StartPartitionedTimeout\"), (117, \"StartFailureTimeout\"), (118, \"HeartbeatIntervalDbDb\"), (119, \"HeartbeatIntervalDbApi\"), (120, \"TimeBetweenLocalCheckpoints\"), (121, \"TimeBetweenGlobalCheckpoints\"), (170, \"TimeBetweenEpochs\"), (171, \"TimeBetweenEpochsTimeout\"), (182, \"MaxBufferedEpochs\"), "
+"(126, \"NoOfFragmentLogFiles\"), (140, \"FragmentLogFileSize\"), (189, \"InitFragmentLogFiles\"), (190, \"DiskIOThreadPool\"), (159, \"MaxNoOfOpenFiles\"), (162, \"InitialNoOfOpenFiles\"), (129, \"TimeBetweenInactiveTransactionAbortCheck\"), (130, \"TransactionInactiveTimeout\"), (131, \"TransactionDeadlockDetectionTimeout\"), (148, \"Diskless\"), (122, \"ArbitrationTimeout\"), (142, \"Arbitration\"), (7, \"DataDir\"), (125, \"FileSystemPath\"), (250, \"LogLevelStartup\"), (251, \"LogLevelShutdown\"), (252, \"LogLevelStatistic\"), (253, \"L"
+"ogLevelCheckpoint\"), (254, \"LogLevelNodeRestart\"), (255, \"LogLevelConnection\"), (259, \"LogLevelCongestion\"), (258, \"LogLevelError\"), (256, \"LogLevelInfo\"), (158, \"BackupDataDir\"), (163, \"DiskSyncSize\"), (164, \"DiskCheckpointSpeed\"), (165, \"DiskCheckpointSpeedInRestart\"), (133, \"BackupMemory\"), (134, \"BackupDataBufferSize\"), (135, \"BackupLogBufferSize\"), (136, \"BackupWriteSize\"), (139, \"BackupMaxWriteSize\"), (161, \"StringMemory\"), (169, \"MaxAllocate\"), (166, \"MemReportFrequency\"), (167, \"BackupReportFrequen"
+"cy\"), (184, \"StartupStatusReportFrequency\"), (168, \"ODirect\"), (172, \"CompressedBackup\"), (173, \"CompressedLCP\"), (9, \"TotalSendBufferMemory\"), (202, \"ReservedSendBufferMemory\"), (185, \"Nodegroup\"), (186, \"MaxNoOfExecutionThreads\"), (188, \"__ndbmt_lqh_workers\"), (187, \"__ndbmt_lqh_threads\"), (191, \"__ndbmt_classic\"), (193, \"FileSystemPathDD\"), (194, \"FileSystemPathDataFiles\"), (195, \"FileSystemPathUndoFiles\"), (196, \"InitialLogfileGroup\"), (197, \"InitialTablespace\"), (605, \"MaxLCPStartDelay\"), (606, \"Build"
+"IndexThreads\"), (607, \"HeartbeatOrder\"), (608, \"DictTrace\"), (609, \"MaxStartFailRetries\"), (610, \"StartFailRetryDelay\")','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.transporters\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`transporters` AS SELECT node_id, remote_node_id,  CASE connection_status  WHEN 0 THEN \"CONNECTED\"  WHEN 1 THEN \"CONNECTING\"  WHEN 2 THEN \"DISCONNECTED\"  WHEN 3 THEN \"DISCONNECTING\"  ELSE NULL  END AS status FROM ndbinfo.ndb$transporters','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.logspaces\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`logspaces` AS SELECT node_id,  CASE log_type  WHEN 0 THEN \"REDO\"  WHEN 1 THEN \"DD-UNDO\"  ELSE NULL  END AS log_type, log_id, log_part, total, used FROM ndbinfo.ndb$logspaces','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.logbuffers\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`logbuffers` AS SELECT node_id,  CASE log_type  WHEN 0 THEN \"REDO\"  WHEN 1 THEN \"DD-UNDO\"  ELSE \"<unknown>\"  END AS log_type, log_id, log_part, total, used FROM ndbinfo.ndb$logbuffers','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.resources\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`resources` AS SELECT node_id,  CASE resource_id  WHEN 0 THEN \"RESERVED\"  WHEN 1 THEN \"DISK_OPERATIONS\"  WHEN 2 THEN \"DISK_RECORDS\"  WHEN 3 THEN \"DATA_MEMORY\"  WHEN 4 THEN \"JOBBUFFER\"  WHEN 5 THEN \"FILE_BUFFERS\"  WHEN 6 THEN \"TRANSPORTER_BUFFERS\"  ELSE \"<unknown>\"  END AS resource_name, reserved, used, max FROM ndbinfo.ndb$resources','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.counters\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`counters` AS SELECT node_id, b.block_name, block_instance, counter_id, CASE counter_id  WHEN 1 THEN \"ATTRINFO\"  WHEN 2 THEN \"TRANSACTIONS\"  WHEN 3 THEN \"COMMITS\"  WHEN 4 THEN \"READS\"  WHEN 5 THEN \"SIMPLE_READS\"  WHEN 6 THEN \"WRITES\"  WHEN 7 THEN \"ABORTS\"  WHEN 8 THEN \"TABLE_SCANS\"  WHEN 9 THEN \"RANGE_SCANS\"  WHEN 10 THEN \"OPERATIONS\"  ELSE \"<unknown>\"  END AS counter_name, val FROM ndbinfo.ndb$counter"
+"s c LEFT JOIN ndbinfo.blocks b ON c.block_number = b.block_number','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.nodes\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`nodes` AS SELECT node_id, uptime, CASE status  WHEN 0 THEN \"NOTHING\"  WHEN 1 THEN \"CMVMI\"  WHEN 2 THEN \"STARTING\"  WHEN 3 THEN \"STARTED\"  WHEN 4 THEN \"SINGLEUSER\"  WHEN 5 THEN \"STOPPING_1\"  WHEN 6 THEN \"STOPPING_2\"  WHEN 7 THEN \"STOPPING_3\"  WHEN 8 THEN \"STOPPING_4\"  ELSE \"<unknown>\"  END AS status, start_phase FROM ndbinfo.ndb$nodes','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.memoryusage\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`memoryusage` AS SELECT node_id,  pool_name AS memory_type,  SUM(used*entry_size) AS used,  SUM(used) AS used_pages,  SUM(total*entry_size) AS total,  SUM(total) AS total_pages FROM ndbinfo.ndb$pools WHERE block_number IN (248, 254) AND   (pool_name = \"Index memory\" OR pool_name = \"Data memory\") GROUP BY node_id, memory_type','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"\n "
+"# ndbinfo.diskpagebuffer\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE SQL SECURITY INVOKER VIEW `ndbinfo`.`diskpagebuffer` AS SELECT node_id, block_instance, pages_written, pages_written_lcp, pages_read, log_waits, page_requests_direct_return, page_requests_wait_queue, page_requests_wait_io FROM ndbinfo.ndb$diskpagebuffer','SET @dummy = 0');\n "
+"PREPARE stmt FROM @str;\n "
+"EXECUTE stmt;\n "
+"DROP PREPARE stmt;\n "
+"# Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.\n "
+"# \n "
+"# This program is free software; you can redistribute it and/or modify\n "
+"# it under the terms of the GNU General Public License as published by\n "
+"# the Free Software Foundation; version 2 of the License.\n "
+"# \n "
+"# This program is distributed in the hope that it will be useful,\n "
+"# but WITHOUT ANY WARRANTY; without even the implied warranty of\n "
+"# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n "
+"# GNU General Public License for more details.\n "
+"# \n "
+"# You should have received a copy of the GNU General Public License\n "
+"# along with this program; if not, write to the Free Software\n "
+"# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA\n "
 "\n "
 "# This part converts any old privilege tables to privilege tables suitable\n "
 "# for current version of MySQL\n "
@@ -166,7 +482,7 @@ const char* mysql_fix_privilege_tables={
 "    COLLATE utf8_general_ci DEFAULT '' NOT NULL,\n "
 "  MODIFY Table_priv set('Select','Insert','Update','Delete','Create',\n "
 "                        'Drop','Grant','References','Index','Alter',\n "
-"                        'Create View','Show view')\n "
+"                        'Create View','Show view','Trigger')\n "
 "    COLLATE utf8_general_ci DEFAULT '' NOT NULL,\n "
 "  COMMENT='Table privileges';\n "
 "\n "
@@ -324,6 +640,37 @@ const char* mysql_fix_privilege_tables={
 "  MODIFY type enum ('function','aggregate') COLLATE utf8_general_ci NOT NULL;\n "
 "\n "
 "#\n "
+"# Modify log tables.\n "
+"#\n "
+"\n "
+"SET @old_log_state = @@global.general_log;\n "
+"SET GLOBAL general_log = 'OFF';\n "
+"ALTER TABLE general_log\n "
+"  MODIFY event_time TIMESTAMP NOT NULL,\n "
+"  MODIFY user_host MEDIUMTEXT NOT NULL,\n "
+"  MODIFY thread_id INTEGER NOT NULL,\n "
+"  MODIFY server_id INTEGER UNSIGNED NOT NULL,\n "
+"  MODIFY command_type VARCHAR(64) NOT NULL,\n "
+"  MODIFY argument MEDIUMTEXT NOT NULL;\n "
+"SET GLOBAL general_log = @old_log_state;\n "
+"\n "
+"SET @old_log_state = @@global.slow_query_log;\n "
+"SET GLOBAL slow_query_log = 'OFF';\n "
+"ALTER TABLE slow_log\n "
+"  MODIFY start_time TIMESTAMP NOT NULL,\n "
+"  MODIFY user_host MEDIUMTEXT NOT NULL,\n "
+"  MODIFY query_time TIME NOT NULL,\n "
+"  MODIFY lock_time TIME NOT NULL,\n "
+"  MODIFY rows_sent INTEGER NOT NULL,\n "
+"  MODIFY rows_examined INTEGER NOT NULL,\n "
+"  MODIFY db VARCHAR(512) NOT NULL,\n "
+"  MODIFY last_insert_id INTEGER NOT NULL,\n "
+"  MODIFY insert_id INTEGER NOT NULL,\n "
+"  MODIFY server_id INTEGER UNSIGNED NOT NULL,\n "
+"  MODIFY sql_text MEDIUMTEXT NOT NULL;\n "
+"SET GLOBAL slow_query_log = @old_log_state;\n "
+"\n "
+"#\n "
 "# Detect whether we had Create_view_priv\n "
 "#\n "
 "SET @hadCreateViewPriv:=0;\n "
@@ -431,6 +778,10 @@ const char* mysql_fix_privilege_tables={
 "  MODIFY Proc_priv set('Execute','Alter Routine','Grant')\n "
 "    COLLATE utf8_general_ci DEFAULT '' NOT NULL;\n "
 "\n "
+"ALTER IGNORE TABLE procs_priv\n "
+"  MODIFY Routine_name char(64)\n "
+"    COLLATE utf8_general_ci DEFAULT '' NOT NULL;\n "
+"\n "
 "ALTER TABLE procs_priv\n "
 "  ADD Routine_type enum('FUNCTION','PROCEDURE')\n "
 "    COLLATE utf8_general_ci NOT NULL AFTER Routine_name;\n "
@@ -505,18 +856,48 @@ const char* mysql_fix_privilege_tables={
 "ALTER TABLE proc MODIFY character_set_client\n "
 "                        char(32) collate utf8_bin DEFAULT NULL;\n "
 "\n "
+"SELECT CASE WHEN COUNT(*) > 0 THEN \n "
+"CONCAT (\"WARNING: NULL values of the 'character_set_client' column ('mysql.proc' table) have been updated with a default value (\", @@character_set_client, \"). Please verify if necessary.\")\n "
+"ELSE NULL \n "
+"END \n "
+"AS value FROM proc WHERE character_set_client IS NULL;\n "
+"\n "
+"UPDATE proc SET character_set_client = @@character_set_client \n "
+"                     WHERE character_set_client IS NULL;\n "
+"\n "
 "ALTER TABLE proc ADD collation_connection\n "
 "                     char(32) collate utf8_bin DEFAULT NULL\n "
 "                     AFTER character_set_client;\n "
 "ALTER TABLE proc MODIFY collation_connection\n "
 "                        char(32) collate utf8_bin DEFAULT NULL;\n "
 "\n "
+"SELECT CASE WHEN COUNT(*) > 0 THEN \n "
+"CONCAT (\"WARNING: NULL values of the 'collation_connection' column ('mysql.proc' table) have been updated with a default value (\", @@collation_connection, \"). Please verify if necessary.\")\n "
+"ELSE NULL \n "
+"END \n "
+"AS value FROM proc WHERE collation_connection IS NULL;\n "
+"\n "
+"UPDATE proc SET collation_connection = @@collation_connection\n "
+"                     WHERE collation_connection IS NULL;\n "
+"\n "
 "ALTER TABLE proc ADD db_collation\n "
 "                     char(32) collate utf8_bin DEFAULT NULL\n "
 "                     AFTER collation_connection;\n "
 "ALTER TABLE proc MODIFY db_collation\n "
 "                        char(32) collate utf8_bin DEFAULT NULL;\n "
 "\n "
+"SELECT CASE WHEN COUNT(*) > 0 THEN \n "
+"CONCAT (\"WARNING: NULL values of the 'db_collation' column ('mysql.proc' table) have been updated with default values. Please verify if necessary.\")\n "
+"ELSE NULL\n "
+"END\n "
+"AS value FROM proc WHERE db_collation IS NULL;\n "
+"\n "
+"UPDATE proc AS p SET db_collation  = \n "
+"                     ( SELECT DEFAULT_COLLATION_NAME \n "
+"                       FROM INFORMATION_SCHEMA.SCHEMATA \n "
+"                       WHERE SCHEMA_NAME = p.db)\n "
+"                     WHERE db_collation IS NULL;\n "
+"\n "
 "ALTER TABLE proc ADD body_utf8 longblob DEFAULT NULL\n "
 "                     AFTER db_collation;\n "
 "ALTER TABLE proc MODIFY body_utf8 longblob DEFAULT NULL;\n "
@@ -541,7 +922,10 @@ const char* mysql_fix_privilege_tables={
 "#\n "
 "ALTER TABLE event DROP PRIMARY KEY;\n "
 "ALTER TABLE event ADD PRIMARY KEY(db, name);\n "
-"ALTER TABLE event ADD sql_mode\n "
+"# Add sql_mode column just in case.\n "
+"ALTER TABLE event ADD sql_mode set ('NOT_USED') AFTER on_completion;\n "
+"# Update list of sql_mode values.\n "
+"ALTER TABLE event MODIFY sql_mode\n "
 "                        set('REAL_AS_FLOAT',\n "
 "                            'PIPES_AS_CONCAT',\n "
 "                            'ANSI_QUOTES',\n "
@@ -576,7 +960,10 @@ const char* mysql_fix_privilege_tables={
 "                            'PAD_CHAR_TO_FULL_LENGTH'\n "
 "                            ) DEFAULT '' NOT NULL AFTER on_completion;\n "
 "ALTER TABLE event MODIFY name char(64) CHARACTER SET utf8 NOT NULL default '';\n "
-"ALTER TABLE event ADD COLUMN originator INT(10) NOT NULL AFTER comment;\n "
+"\n "
+"ALTER TABLE event MODIFY COLUMN originator INT UNSIGNED NOT NULL;\n "
+"ALTER TABLE event ADD COLUMN originator INT UNSIGNED NOT NULL AFTER comment;\n "
+"\n "
 "ALTER TABLE event MODIFY COLUMN status ENUM('ENABLED','DISABLED','SLAVESIDE_DISABLED') NOT NULL default 'ENABLED';\n "
 "\n "
 "ALTER TABLE event ADD COLUMN time_zone char(64) CHARACTER SET latin1\n "
@@ -621,8 +1008,6 @@ const char* mysql_fix_privilege_tables={
 "ALTER TABLE db ADD Trigger_priv enum('N','Y') COLLATE utf8_general_ci DEFAULT 'N' NOT NULL;\n "
 "ALTER TABLE db MODIFY Trigger_priv enum('N','Y') COLLATE utf8_general_ci DEFAULT 'N' NOT NULL;\n "
 "\n "
-"ALTER TABLE tables_priv MODIFY Table_priv set('Select','Insert','Update','Delete','Create','Drop','Grant','References','Index','Alter','Create View','Show view','Trigger') COLLATE utf8_general_ci DEFAULT '' NOT NULL;\n "
-"\n "
 "UPDATE user SET Trigger_priv=Super_priv WHERE @hadTriggerPriv = 0;\n "
 "\n "
 "# Activate the new, possible modified privilege tables\n "
@@ -630,5 +1015,13 @@ const char* mysql_fix_privilege_tables={
 "# changes was correct\n "
 "\n "
 "flush privileges;\n "
+"\n "
+"#\n "
+"# ndb_binlog_index table\n "
+"#\n "
+"ALTER TABLE ndb_binlog_index\n "
+"  ADD COLUMN next_position BIGINT UNSIGNED NOT NULL;\n "
+"ALTER TABLE ndb_binlog_index\n "
+"  ADD COLUMN next_file VARCHAR(255) NOT NULL;"
 "\
 "};
