@@ -1,6 +1,5 @@
 /*
-   Copyright (c) 2000-2003, 2007, 2008 MySQL AB, 2008 Sun Microsystems, Inc.
-   Use is subject to license terms.
+   Copyright (c) 2000, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,16 +14,18 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */
-int ndbcluster_connect(int (*connect_callback)(void));
-int ndbcluster_disconnect();
+int ndbcluster_connect(int (*connect_callback)(void),
+                       ulong wait_connected,
+                       uint connection_pool_size,
+                       bool optimized_node_select,
+                       const char* connect_string, uint force_nodeid);
+void ndbcluster_disconnect(void);
 
 Ndb_cluster_connection *ndb_get_cluster_connection();
 ulonglong ndb_get_latest_trans_gci();
 void ndb_set_latest_trans_gci(ulonglong val);
 int ndb_has_node_id(uint id);
-
-/* options from from mysqld.cc */
-extern ulong opt_ndb_cluster_connection_pool;
+void ndb_get_connection_stats(Uint64* statsArr);
 
 /* perform random sleep in the range milli_sleep to 2*milli_sleep */
 inline void do_retry_sleep(unsigned milli_sleep)
