@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,24 +25,32 @@
  * For historical reasons it looks like we are adding random fragments
  * and attributes to existing index.  In fact all fragments must be
  * created at one time and they have identical attributes.
+ *
+ * But history changes?
+ * Now index will be created using the sequence
+ *   CREATE_TAB_REQ
+ *     TUP_ADD_ATTR_REQ +
+ *
+ * Followed by 0-N
+ *   TUXFRAGREQ
  */
 
+#include <signaldata/CreateTab.hpp>
+#include <signaldata/LqhFrag.hpp>
+
 void
-Dbtux::execTUXFRAGREQ(Signal* signal)
+Dbtux::execCREATE_TAB_REQ(Signal* signal)
 {
   jamEntry();
-  if (signal->theData[0] == (Uint32)-1) {
-    jam();
-    abortAddFragOp(signal);
-    return;
-  }
-  const TuxFragReq reqCopy = *(const TuxFragReq*)signal->getDataPtr();
-  const TuxFragReq* const req = &reqCopy;
+  CreateTabReq copy = *(CreateTabReq*)signal->getDataPtr();
+  CreateTabReq* req = &copy;
+
   IndexPtr indexPtr;
   indexPtr.i = RNIL;
   FragOpPtr fragOpPtr;
   fragOpPtr.i = RNIL;
-  TuxFragRef::ErrorCode errorCode = TuxFragRef::NoError;
+  Uint32 errorCode = 0;
+
   do {
     // get the index record
     if (req->tableId >= c_indexPool.getSize()) {
@@ -51,22 +59,23 @@ Dbtux::execTUXFRAGREQ(Signal* signal)
       break;
     }
     c_indexPool.getPtr(indexPtr, req->tableId);
-    if (indexPtr.p->m_state != Index::NotDefined &&
-        indexPtr.p->m_state != Index::Defining) {
+    if (indexPtr.p->m_state != Index::NotDefined)
+    {
       jam();
       errorCode = TuxFragRef::InvalidRequest;
       indexPtr.i = RNIL;        // leave alone
       break;
     }
+
     // get new operation record
     c_fragOpPool.seize(fragOpPtr);
     ndbrequire(fragOpPtr.i != RNIL);
     new (fragOpPtr.p) FragOp();
-    fragOpPtr.p->m_userPtr = req->userPtr;
-    fragOpPtr.p->m_userRef = req->userRef;
+    fragOpPtr.p->m_userPtr = req->senderData;
+    fragOpPtr.p->m_userRef = req->senderRef;
     fragOpPtr.p->m_indexId = req->tableId;
-    fragOpPtr.p->m_fragId = req->fragId;
-    fragOpPtr.p->m_fragNo = indexPtr.p->m_numFrags;
+    fragOpPtr.p->m_fragId = RNIL;
+    fragOpPtr.p->m_fragNo = RNIL;
     fragOpPtr.p->m_numAttrsRecvd = 0;
 #ifdef VM_TRACE
     if (debugFlags & DebugMeta) {
@@ -74,68 +83,24 @@ Dbtux::execTUXFRAGREQ(Signal* signal)
     }
 #endif
     // check if index has place for more fragments
-    ndbrequire(indexPtr.p->m_numFrags < MaxIndexFragments);
-    // seize new fragment record
-    FragPtr fragPtr;
-    c_fragPool.seize(fragPtr);
-    if (fragPtr.i == RNIL) {
+    ndbrequire(indexPtr.p->m_state == Index::NotDefined &&
+               DictTabInfo::isOrderedIndex(req->tableType) &&
+               req->noOfAttributes > 0 &&
+               req->noOfAttributes <= MaxIndexAttributes &&
+               indexPtr.p->m_descPage == RNIL);
+
+    indexPtr.p->m_state = Index::Defining;
+    indexPtr.p->m_tableType = (DictTabInfo::TableType)req->tableType;
+    indexPtr.p->m_tableId = req->primaryTableId;
+    indexPtr.p->m_numAttrs = req->noOfAttributes;
+    indexPtr.p->m_storeNullKey = true;  // not yet configurable
+    // allocate attribute descriptors
+    if (! allocDescEnt(indexPtr)) {
       jam();
-      errorCode = TuxFragRef::NoFreeFragment;
+      errorCode = TuxFragRef::NoFreeAttributes;
       break;
     }
-    new (fragPtr.p) Frag(c_scanOpPool);
-    fragPtr.p->m_tableId = req->primaryTableId;
-    fragPtr.p->m_indexId = req->tableId;
-    fragPtr.p->m_fragId = req->fragId;
-    fragPtr.p->m_numAttrs = req->noOfAttr;
-    fragPtr.p->m_storeNullKey = true;  // not yet configurable
-    fragPtr.p->m_tupIndexFragPtrI = req->tupIndexFragPtrI;
-    fragPtr.p->m_tupTableFragPtrI = req->tupTableFragPtrI[0];
-    fragPtr.p->m_accTableFragPtrI = req->accTableFragPtrI[0];
-    // add the fragment to the index
-    indexPtr.p->m_fragId[indexPtr.p->m_numFrags] = req->fragId;
-    indexPtr.p->m_fragPtrI[indexPtr.p->m_numFrags] = fragPtr.i;
-    indexPtr.p->m_numFrags++;
-    // save under operation
-    fragOpPtr.p->m_fragPtrI = fragPtr.i;
-    // prepare to receive attributes
-    if (fragOpPtr.p->m_fragNo == 0) {
-      jam();
-      // receiving first fragment
-      ndbrequire(
-          indexPtr.p->m_state == Index::NotDefined &&
-          DictTabInfo::isOrderedIndex(req->tableType) &&
-          req->noOfAttr > 0 &&
-          req->noOfAttr <= MaxIndexAttributes &&
-          indexPtr.p->m_descPage == RNIL);
-      indexPtr.p->m_state = Index::Defining;
-      indexPtr.p->m_tableType = (DictTabInfo::TableType)req->tableType;
-      indexPtr.p->m_tableId = req->primaryTableId;
-      indexPtr.p->m_numAttrs = req->noOfAttr;
-      indexPtr.p->m_storeNullKey = true;  // not yet configurable
-      // allocate attribute descriptors
-      if (! allocDescEnt(indexPtr)) {
-        jam();
-        errorCode = TuxFragRef::NoFreeAttributes;
-        break;
-      }
-    } else {
-      // receiving subsequent fragment
-      jam();
-      ndbrequire(
-          indexPtr.p->m_state == Index::Defining &&
-          indexPtr.p->m_tableType == (DictTabInfo::TableType)req->tableType &&
-          indexPtr.p->m_tableId == req->primaryTableId &&
-          indexPtr.p->m_numAttrs == req->noOfAttr);
-    }
-    // copy metadata address to each fragment
-    fragPtr.p->m_descPage = indexPtr.p->m_descPage;
-    fragPtr.p->m_descOff = indexPtr.p->m_descOff;
-#ifdef VM_TRACE
-    if (debugFlags & DebugMeta) {
-      debugOut << "Add frag " << fragPtr.i << " " << *fragPtr.p << endl;
-    }
-#endif
+
     // error inserts
     if ((ERROR_INSERTED(12001) && fragOpPtr.p->m_fragNo == 0) ||
         (ERROR_INSERTED(12002) && fragOpPtr.p->m_fragNo == 1)) {
@@ -145,33 +110,32 @@ Dbtux::execTUXFRAGREQ(Signal* signal)
       break;
     }
     // success
-    TuxFragConf* const conf = (TuxFragConf*)signal->getDataPtrSend();
-    conf->userPtr = req->userPtr;
+    CreateTabConf* conf = (CreateTabConf*)signal->getDataPtrSend();
+    conf->senderRef = reference();
+    conf->senderData = req->senderData;
     conf->tuxConnectPtr = fragOpPtr.i;
-    conf->fragPtr = fragPtr.i;
-    conf->fragId = fragPtr.p->m_fragId;
-    sendSignal(req->userRef, GSN_TUXFRAGCONF,
-        signal, TuxFragConf::SignalLength, JBB);
+    sendSignal(req->senderRef, GSN_CREATE_TAB_CONF,
+               signal, CreateTabConf::SignalLength, JBB);
     return;
   } while (0);
   // error
-  TuxFragRef* const ref = (TuxFragRef*)signal->getDataPtrSend();
-  ref->userPtr = req->userPtr;
+
+  CreateTabRef* const ref = (CreateTabRef*)signal->getDataPtrSend();
+  ref->senderData = req->senderData;
   ref->errorCode = errorCode;
-  sendSignal(req->userRef, GSN_TUXFRAGREF,
-      signal, TuxFragRef::SignalLength, JBB);
-  if (fragOpPtr.i != RNIL) {
-#ifdef VM_TRACE
-    if (debugFlags & DebugMeta) {
-      debugOut << "Release on frag error frag op " << fragOpPtr.i << " " << *fragOpPtr.p << endl;
-    }
-#endif
-    c_fragOpPool.release(fragOpPtr);
-  }
+  sendSignal(req->senderRef, GSN_CREATE_TAB_REF,
+             signal, CreateTabRef::SignalLength, JBB);
+
   if (indexPtr.i != RNIL) {
     jam();
     // let DICT drop the unfinished index
   }
+
+  if (fragOpPtr.i != RNIL)
+  {
+    jam();
+    c_fragOpPool.release(fragOpPtr);
+  }
 }
 
 void
@@ -183,10 +147,8 @@ Dbtux::execTUX_ADD_ATTRREQ(Signal* signal)
   // get the records
   FragOpPtr fragOpPtr;
   IndexPtr indexPtr;
-  FragPtr fragPtr;
   c_fragOpPool.getPtr(fragOpPtr, req->tuxConnectPtr);
   c_indexPool.getPtr(indexPtr, fragOpPtr.p->m_indexId);
-  c_fragPool.getPtr(fragPtr, fragOpPtr.p->m_fragPtrI);
   TuxAddAttrRef::ErrorCode errorCode = TuxAddAttrRef::NoError;
   do {
     // expected attribute id
@@ -195,87 +157,74 @@ Dbtux::execTUX_ADD_ATTRREQ(Signal* signal)
         indexPtr.p->m_state == Index::Defining &&
         attrId < indexPtr.p->m_numAttrs &&
         attrId == req->attrId);
-    // define the attribute
-    DescEnt& descEnt = getDescEnt(indexPtr.p->m_descPage, indexPtr.p->m_descOff);
-    DescAttr& descAttr = descEnt.m_descAttr[attrId];
-    descAttr.m_attrDesc = req->attrDescriptor;
-    descAttr.m_primaryAttrId = req->primaryAttrId;
-    descAttr.m_typeId = AttributeDescriptor::getType(req->attrDescriptor);
-    descAttr.m_charset = (req->extTypeInfo >> 16);
-#ifdef VM_TRACE
-    if (debugFlags & DebugMeta) {
-      debugOut << "Add frag " << fragPtr.i << " attr " << attrId << " " << descAttr << endl;
-    }
-#endif
-    // check that type is valid and has a binary comparison method
-    const NdbSqlUtil::Type& type = NdbSqlUtil::getTypeBinary(descAttr.m_typeId);
-    if (type.m_typeId == NdbSqlUtil::Type::Undefined ||
-        type.m_cmp == 0) {
+    const Uint32 ad = req->attrDescriptor;
+    const Uint32 typeId = AttributeDescriptor::getType(ad);
+    const Uint32 sizeInBytes = AttributeDescriptor::getSizeInBytes(ad);
+    const Uint32 nullable = AttributeDescriptor::getNullable(ad);
+    const Uint32 csNumber = req->extTypeInfo >> 16;
+    const Uint32 primaryAttrId = req->primaryAttrId;
+
+    DescHead& descHead = getDescHead(*indexPtr.p);
+    // add type to spec
+    KeySpec& keySpec = indexPtr.p->m_keySpec;
+    KeyType keyType(typeId, sizeInBytes, nullable, csNumber);
+    if (keySpec.add(keyType) == -1) {
       jam();
       errorCode = TuxAddAttrRef::InvalidAttributeType;
       break;
     }
-    if (descAttr.m_charset != 0) {
-      uint err;
-      CHARSET_INFO *cs = all_charsets[descAttr.m_charset];
+    // add primary attr to read keys array
+    AttributeHeader* keyAttrs = getKeyAttrs(descHead);
+    AttributeHeader& keyAttr = keyAttrs[attrId];
+    new (&keyAttr) AttributeHeader(primaryAttrId, sizeInBytes);
+#ifdef VM_TRACE
+    if (debugFlags & DebugMeta) {
+      debugOut << "attr " << attrId << " " << keyType << endl;
+    }
+#endif
+    if (csNumber != 0) {
+      unsigned err;
+      CHARSET_INFO *cs = all_charsets[csNumber];
       ndbrequire(cs != 0);
-      if ((err = NdbSqlUtil::check_column_for_ordered_index(descAttr.m_typeId, cs))) {
+      if ((err = NdbSqlUtil::check_column_for_ordered_index(typeId, cs))) {
         jam();
         errorCode = (TuxAddAttrRef::ErrorCode) err;
         break;
       }
     }
     const bool lastAttr = (indexPtr.p->m_numAttrs == fragOpPtr.p->m_numAttrsRecvd);
-    if ((ERROR_INSERTED(12003) && fragOpPtr.p->m_fragNo == 0 && attrId == 0) ||
-        (ERROR_INSERTED(12004) && fragOpPtr.p->m_fragNo == 0 && lastAttr) ||
-        (ERROR_INSERTED(12005) && fragOpPtr.p->m_fragNo == 1 && attrId == 0) ||
-        (ERROR_INSERTED(12006) && fragOpPtr.p->m_fragNo == 1 && lastAttr)) {
+    if ((ERROR_INSERTED(12003) && attrId == 0) ||
+        (ERROR_INSERTED(12004) && lastAttr))
+    {
       errorCode = (TuxAddAttrRef::ErrorCode)1;
       CLEAR_ERROR_INSERT_VALUE;
       break;
     }
     if (lastAttr) {
-      jam();
-      // initialize tree header
-      TreeHead& tree = fragPtr.p->m_tree;
-      new (&tree) TreeHead();
-      // make these configurable later
-      tree.m_nodeSize = MAX_TTREE_NODE_SIZE;
-      tree.m_prefSize = MAX_TTREE_PREF_SIZE;
-      const unsigned maxSlack = MAX_TTREE_NODE_SLACK;
-      // size up to and including first 2 entries
-      const unsigned pref = tree.getSize(AccPref);
-      if (! (pref <= tree.m_nodeSize)) {
-        jam();
-        errorCode = TuxAddAttrRef::InvalidNodeSize;
-        break;
-      }
-      const unsigned slots = (tree.m_nodeSize - pref) / TreeEntSize;
-      // leave out work space entry
-      tree.m_maxOccup = 2 + slots - 1;
-      // min occupancy of interior node must be at least 2
-      if (! (2 + maxSlack <= tree.m_maxOccup)) {
-        jam();
-        errorCode = TuxAddAttrRef::InvalidNodeSize;
-        break;
-      }
-      tree.m_minOccup = tree.m_maxOccup - maxSlack;
-      // root node does not exist (also set by ctor)
-      tree.m_root = NullTupLoc;
+      // compute min prefix
+      const KeySpec& keySpec = indexPtr.p->m_keySpec;
+      unsigned attrs = 0;
+      unsigned bytes = keySpec.get_nullmask_len(false);
+      unsigned maxAttrs = indexPtr.p->m_numAttrs;
 #ifdef VM_TRACE
-      if (debugFlags & DebugMeta) {
-        if (fragOpPtr.p->m_fragNo == 0) {
-          debugOut << "Index id=" << indexPtr.i;
-          debugOut << " nodeSize=" << tree.m_nodeSize;
-          debugOut << " headSize=" << NodeHeadSize;
-          debugOut << " prefSize=" << tree.m_prefSize;
-          debugOut << " entrySize=" << TreeEntSize;
-          debugOut << " minOccup=" << tree.m_minOccup;
-          debugOut << " maxOccup=" << tree.m_maxOccup;
-          debugOut << endl;
-        }
+      {
+        const char* p = NdbEnv_GetEnv("MAX_TTREE_PREF_ATTRS", (char*)0, 0);
+        if (p != 0 && p[0] != 0 && maxAttrs > (unsigned)atoi(p))
+          maxAttrs = atoi(p);
       }
 #endif
+      while (attrs < maxAttrs) {
+        const KeyType& keyType = keySpec.get_type(attrs);
+        const unsigned newbytes = bytes + keyType.get_byte_size();
+        if (newbytes > (MAX_TTREE_PREF_SIZE << 2))
+          break;
+        attrs++;
+        bytes = newbytes;
+      }
+      if (attrs == 0)
+        bytes = 0;
+      indexPtr.p->m_prefAttrs = attrs;
+      indexPtr.p->m_prefBytes = bytes;
       // fragment is defined
 #ifdef VM_TRACE
       if (debugFlags & DebugMeta) {
@@ -303,10 +252,147 @@ Dbtux::execTUX_ADD_ATTRREQ(Signal* signal)
       debugOut << "Release on attr error frag op " << fragOpPtr.i << " " << *fragOpPtr.p << endl;
     }
 #endif
-  c_fragOpPool.release(fragOpPtr);
   // let DICT drop the unfinished index
 }
 
+void
+Dbtux::execTUXFRAGREQ(Signal* signal)
+{
+  jamEntry();
+
+  if (signal->theData[0] == (Uint32)-1) {
+    jam();
+    abortAddFragOp(signal);
+    return;
+  }
+
+  const TuxFragReq reqCopy = *(const TuxFragReq*)signal->getDataPtr();
+  const TuxFragReq* const req = &reqCopy;
+  IndexPtr indexPtr;
+  indexPtr.i = RNIL;
+  TuxFragRef::ErrorCode errorCode = TuxFragRef::NoError;
+  do {
+    // get the index record
+    if (req->tableId >= c_indexPool.getSize()) {
+      jam();
+      errorCode = TuxFragRef::InvalidRequest;
+      break;
+    }
+    c_indexPool.getPtr(indexPtr, req->tableId);
+    if (false && indexPtr.p->m_state != Index::Defining) {
+      jam();
+      errorCode = TuxFragRef::InvalidRequest;
+      indexPtr.i = RNIL;        // leave alone
+      break;
+    }
+
+    // check if index has place for more fragments
+    ndbrequire(indexPtr.p->m_numFrags < MaxIndexFragments);
+    // seize new fragment record
+    if (ERROR_INSERTED(12008))
+    {
+      CLEAR_ERROR_INSERT_VALUE;
+      errorCode = TuxFragRef::InvalidRequest;
+      break;
+    }
+
+    FragPtr fragPtr;
+    c_fragPool.seize(fragPtr);
+    if (fragPtr.i == RNIL) {
+      jam();
+      errorCode = TuxFragRef::NoFreeFragment;
+      break;
+    }
+    new (fragPtr.p) Frag(c_scanOpPool);
+    fragPtr.p->m_tableId = req->primaryTableId;
+    fragPtr.p->m_indexId = req->tableId;
+    fragPtr.p->m_fragId = req->fragId;
+    fragPtr.p->m_tupIndexFragPtrI = req->tupIndexFragPtrI;
+    fragPtr.p->m_tupTableFragPtrI = req->tupTableFragPtrI;
+    fragPtr.p->m_accTableFragPtrI = req->accTableFragPtrI;
+    // add the fragment to the index
+    Uint32 fragNo = indexPtr.p->m_numFrags;
+    indexPtr.p->m_fragId[indexPtr.p->m_numFrags] = req->fragId;
+    indexPtr.p->m_fragPtrI[indexPtr.p->m_numFrags] = fragPtr.i;
+    indexPtr.p->m_numFrags++;
+#ifdef VM_TRACE
+    if (debugFlags & DebugMeta) {
+      debugOut << "Add frag " << fragPtr.i << " " << *fragPtr.p << endl;
+    }
+#endif
+    // error inserts
+    if ((ERROR_INSERTED(12001) && fragNo == 0) ||
+        (ERROR_INSERTED(12002) && fragNo == 1)) {
+      jam();
+      errorCode = (TuxFragRef::ErrorCode)1;
+      CLEAR_ERROR_INSERT_VALUE;
+      break;
+    }
+
+    // initialize tree header
+    TreeHead& tree = fragPtr.p->m_tree;
+    new (&tree) TreeHead();
+    // make these configurable later
+    tree.m_nodeSize = MAX_TTREE_NODE_SIZE;
+    tree.m_prefSize = (indexPtr.p->m_prefBytes + 3) / 4;
+    const unsigned maxSlack = MAX_TTREE_NODE_SLACK;
+    // size of header and min prefix
+    const unsigned fixedSize = NodeHeadSize + tree.m_prefSize;
+    if (! (fixedSize <= tree.m_nodeSize)) {
+      jam();
+      errorCode = (TuxFragRef::ErrorCode)TuxAddAttrRef::InvalidNodeSize;
+      break;
+    }
+    const unsigned slots = (tree.m_nodeSize - fixedSize) / TreeEntSize;
+    tree.m_maxOccup = slots;
+    // min occupancy of interior node must be at least 2
+    if (! (2 + maxSlack <= tree.m_maxOccup)) {
+      jam();
+      errorCode = (TuxFragRef::ErrorCode)TuxAddAttrRef::InvalidNodeSize;
+      break;
+    }
+    tree.m_minOccup = tree.m_maxOccup - maxSlack;
+    // root node does not exist (also set by ctor)
+    tree.m_root = NullTupLoc;
+#ifdef VM_TRACE
+    if (debugFlags & DebugMeta) {
+      if (fragNo == 0) {
+        debugOut << "Index id=" << indexPtr.i;
+        debugOut << " nodeSize=" << tree.m_nodeSize;
+        debugOut << " headSize=" << NodeHeadSize;
+        debugOut << " prefSize=" << tree.m_prefSize;
+        debugOut << " entrySize=" << TreeEntSize;
+        debugOut << " minOccup=" << tree.m_minOccup;
+        debugOut << " maxOccup=" << tree.m_maxOccup;
+        debugOut << endl;
+      }
+    }
+#endif
+
+    // success
+    TuxFragConf* const conf = (TuxFragConf*)signal->getDataPtrSend();
+    conf->userPtr = req->userPtr;
+    conf->tuxConnectPtr = RNIL;
+    conf->fragPtr = fragPtr.i;
+    conf->fragId = fragPtr.p->m_fragId;
+    sendSignal(req->userRef, GSN_TUXFRAGCONF,
+        signal, TuxFragConf::SignalLength, JBB);
+    return;
+  } while (0);
+
+  // error
+  TuxFragRef* const ref = (TuxFragRef*)signal->getDataPtrSend();
+  ref->userPtr = req->userPtr;
+  ref->errorCode = errorCode;
+  sendSignal(req->userRef, GSN_TUXFRAGREF,
+      signal, TuxFragRef::SignalLength, JBB);
+
+  if (indexPtr.i != RNIL) {
+    jam();
+    // let DICT drop the unfinished index
+  }
+}
+
 /*
  * LQH aborts on-going create index operation.
  */
@@ -331,24 +417,27 @@ Dbtux::abortAddFragOp(Signal* signal)
  * build and is therefore not correct.
  */
 void
-Dbtux::execALTER_INDX_REQ(Signal* signal)
+Dbtux::execALTER_INDX_IMPL_REQ(Signal* signal)
 {
   jamEntry();
-  const AlterIndxReq reqCopy = *(const AlterIndxReq*)signal->getDataPtr();
-  const AlterIndxReq* const req = &reqCopy;
-  // set index online after build
-  IndexPtr indexPtr;
-  c_indexPool.getPtr(indexPtr, req->getIndexId());
+  const AlterIndxImplReq reqCopy = *(const AlterIndxImplReq*)signal->getDataPtr();
+  const AlterIndxImplReq* const req = &reqCopy;
 
-  if (refToBlock(req->getUserRef()) != DBDICT)
+  IndexPtr indexPtr;
+  c_indexPool.getPtr(indexPtr, req->indexId);
+
+  //Uint32 save = indexPtr.p->m_state;
+  if (! (refToBlock(req->senderRef) == DBDICT) &&
+      ! (isNdbMt() && refToMain(req->senderRef) == DBTUX && 
+         refToInstance(req->senderRef) == 0))
   {
     /**
      * DICT has a really distorted view of the world...
      *   ignore it :(
      */
     jam();
-    switch(req->getOnline()){
-    case 0:
+    switch(req->requestType){
+    case AlterIndxImplReq::AlterIndexOffline:
       jam();
       /*
        * This happens at failed index build, and before dropping an
@@ -356,41 +445,32 @@ Dbtux::execALTER_INDX_REQ(Signal* signal)
        */
       indexPtr.p->m_state = Index::Dropping;
       break;
-    case 2:
+    case AlterIndxImplReq::AlterIndexBuilding:
       jam();
       indexPtr.p->m_state = Index::Building;
       break;
     default:
       jam(); // fall-through
-    case 1:
+    case AlterIndxImplReq::AlterIndexOnline:
       jam();
       indexPtr.p->m_state = Index::Online;
       break;
     }
-#ifdef VM_TRACE
-    if (debugFlags & DebugMeta) {
-      debugOut << "Online index " << indexPtr.i << " " << *indexPtr.p << endl;
-    }
-#endif
   }
-
+  
   // success
-  AlterIndxConf* const conf = (AlterIndxConf*)signal->getDataPtrSend();
-  conf->setUserRef(reference());
-  conf->setConnectionPtr(req->getConnectionPtr());
-  conf->setRequestType(req->getRequestType());
-  conf->setTableId(req->getTableId());
-  conf->setIndexId(req->getIndexId());
-  conf->setIndexVersion(req->getIndexVersion());
-  if (req->getUserRef() != 0)
+  AlterIndxImplConf* const conf = (AlterIndxImplConf*)signal->getDataPtrSend();
+  conf->senderRef = reference();
+  conf->senderData = req->senderData;
+  if (req->senderRef != 0)
   {
     /**
      * TUP cheats and does execute direct
      *   setting UserRef to 0
      */
     jam();
-    sendSignal(req->getUserRef(), GSN_ALTER_INDX_CONF,
-               signal, AlterIndxConf::SignalLength, JBB);
+    sendSignal(req->senderRef, GSN_ALTER_INDX_IMPL_CONF,
+               signal, AlterIndxImplConf::SignalLength, JBB);
   }
 }
 
@@ -485,7 +565,7 @@ bool
 Dbtux::allocDescEnt(IndexPtr indexPtr)
 {
   jam();
-  const unsigned size = DescHeadSize + indexPtr.p->m_numAttrs * DescAttrSize;
+  const Uint32 size = getDescSize(*indexPtr.p);
   DescPagePtr pagePtr;
   pagePtr.i = c_descPageList;
   while (pagePtr.i != RNIL) {
@@ -513,9 +593,13 @@ Dbtux::allocDescEnt(IndexPtr indexPtr)
   indexPtr.p->m_descPage = pagePtr.i;
   indexPtr.p->m_descOff = DescPageSize - pagePtr.p->m_numFree;
   pagePtr.p->m_numFree -= size;
-  DescEnt& descEnt = getDescEnt(indexPtr.p->m_descPage, indexPtr.p->m_descOff);
-  descEnt.m_descHead.m_indexId = indexPtr.i;
-  descEnt.m_descHead.pad1 = 0;
+  DescHead& descHead = *(DescHead*)&pagePtr.p->m_data[indexPtr.p->m_descOff];
+  descHead.m_indexId = indexPtr.i;
+  descHead.m_numAttrs = indexPtr.p->m_numAttrs;
+  descHead.m_magic = DescHead::Magic;
+  KeySpec& keySpec = indexPtr.p->m_keySpec;
+  KeyType* keyTypes = getKeyTypes(descHead);
+  keySpec.set_buf(keyTypes, indexPtr.p->m_numAttrs);
   return true;
 }
 
@@ -525,37 +609,77 @@ Dbtux::freeDescEnt(IndexPtr indexPtr)
   DescPagePtr pagePtr;
   c_descPagePool.getPtr(pagePtr, indexPtr.p->m_descPage);
   Uint32* const data = pagePtr.p->m_data;
-  const unsigned size = DescHeadSize + indexPtr.p->m_numAttrs * DescAttrSize;
-  unsigned off = indexPtr.p->m_descOff;
+  const Uint32 size = getDescSize(*indexPtr.p);
+  Uint32 off = indexPtr.p->m_descOff;
   // move the gap to the free area at the top
   while (off + size < DescPageSize - pagePtr.p->m_numFree) {
     jam();
     // next entry to move over the gap
-    DescEnt& descEnt2 = *(DescEnt*)&data[off + size];
-    Uint32 indexId2 = descEnt2.m_descHead.m_indexId;
+    DescHead& descHead2 = *(DescHead*)&data[off + size];
+    Uint32 indexId2 = descHead2.m_indexId;
     Index& index2 = *c_indexPool.getPtr(indexId2);
-    unsigned size2 = DescHeadSize + index2.m_numAttrs * DescAttrSize;
+    Uint32 size2 = getDescSize(index2);
     ndbrequire(
         index2.m_descPage == pagePtr.i &&
-        index2.m_descOff == off + size);
+        index2.m_descOff == off + size &&
+        index2.m_numAttrs == descHead2.m_numAttrs);
     // move the entry (overlapping copy if size < size2)
-    unsigned i;
+    Uint32 i;
     for (i = 0; i < size2; i++) {
       jam();
       data[off + i] = data[off + size + i];
     }
     off += size2;
-    // adjust page offset in index and all fragments
+    // adjust page offset in index
     index2.m_descOff -= size;
-    for (i = 0; i < index2.m_numFrags; i++) {
-      jam();
-      Frag& frag2 = *c_fragPool.getPtr(index2.m_fragPtrI[i]);
-      frag2.m_descOff -= size;
-      ndbrequire(
-          frag2.m_descPage == index2.m_descPage &&
-          frag2.m_descOff == index2.m_descOff);
-    }
+    {
+      // move KeySpec pointer
+      DescHead& descHead2 = getDescHead(index2);
+      KeyType* keyType2 = getKeyTypes(descHead2);
+      index2.m_keySpec.set_buf(keyType2);
+      ndbrequire(index2.m_keySpec.validate() == 0);
+     }
   }
   ndbrequire(off + size == DescPageSize - pagePtr.p->m_numFree);
   pagePtr.p->m_numFree += size;
 }
+
+void
+Dbtux::execDROP_FRAG_REQ(Signal* signal)
+{
+  DropFragReq copy = *(DropFragReq*)signal->getDataPtr();
+  DropFragReq *req = &copy;
+
+  IndexPtr indexPtr;
+  c_indexPool.getPtr(indexPtr, req->tableId);
+  Uint32 i = 0;
+  for (i = 0; i < indexPtr.p->m_numFrags; i++)
+  {
+    jam();
+    if (indexPtr.p->m_fragId[i] == req->fragId)
+    {
+      jam();
+      FragPtr fragPtr;
+      c_fragPool.getPtr(fragPtr, indexPtr.p->m_fragPtrI[i]);
+      c_fragPool.release(fragPtr);
+
+      for (i++; i < indexPtr.p->m_numFrags; i++)
+      {
+        jam();
+        indexPtr.p->m_fragPtrI[i-1] = indexPtr.p->m_fragPtrI[i];
+        indexPtr.p->m_fragId[i-1] = indexPtr.p->m_fragId[i];
+      }
+      indexPtr.p->m_numFrags--;
+      break;
+    }
+  }
+
+
+  // reply to sender
+  DropFragConf* const conf = (DropFragConf*)signal->getDataPtrSend();
+  conf->senderRef = reference();
+  conf->senderData = req->senderData;
+  conf->tableId = req->tableId;
+  sendSignal(req->senderRef, GSN_DROP_FRAG_CONF,
+             signal, DropFragConf::SignalLength, JBB);
+}
