@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,16 +18,7 @@
 
 #include <ndb_global.h>
 
-#include "NdbApiSignal.hpp"
-#include "NdbImpl.hpp"
-#include <NdbTransaction.hpp>
-#include <NdbOperation.hpp>
-#include <NdbIndexOperation.hpp>
-#include <NdbScanOperation.hpp>
-#include <NdbRecAttr.hpp>
-#include <NdbReceiver.hpp>
 #include "API.hpp"
-#include "NdbEventOperationImpl.hpp"
 
 #include <signaldata/TcCommit.hpp>
 #include <signaldata/TcKeyFailConf.hpp>
@@ -40,6 +31,9 @@
 #include <signaldata/ScanFrag.hpp>
 #include <signaldata/ScanTab.hpp>
 #include <signaldata/SumaImpl.hpp>
+#include <signaldata/NodeFailRep.hpp>
+#include <signaldata/NFCompleteRep.hpp>
+#include <signaldata/AllocNodeId.hpp>
 
 #include <ndb_limits.h>
 #include <NdbOut.hpp>
@@ -79,19 +73,27 @@ Ndb::init(int aMaxNoOfTransactions)
   }//if
   theInitState = StartingInit;
   TransporterFacade * theFacade =  theImpl->m_transporter_facade;
-  theFacade->lock_mutex();
-  
-  const int tBlockNo = theFacade->open(this,
-                                       executeMessage, 
-                                       statusMessage);  
-  if ( tBlockNo == -1 ) {
+  theEventBuffer->m_mutex = theFacade->theMutexPtr;
+
+  const Uint32 tRef = theImpl->open(theFacade);
+
+  if (tRef == 0)
+  {
     theError.code = 4105;
-    theFacade->unlock_mutex();
     DBUG_RETURN(-1); // no more free blocknumbers
   }//if
   
-  theNdbBlockNumber = tBlockNo;
+  Uint32 nodeId = refToNode(tRef);
+  theNdbBlockNumber = refToBlock(tRef);
 
+  if (nodeId > 0)
+  {
+    connected(Uint32(tRef));
+  }
+
+  /* Init cached min node version */
+  theFacade->lock_mutex();
+  theCachedMinDbNodeVersion = theFacade->getMinDbNodeVersion();
   theFacade->unlock_mutex();
   
   theDictionary->setTransporter(this, theFacade);
@@ -140,13 +142,14 @@ Ndb::init(int aMaxNoOfTransactions)
   for (i = 0; i < 16; i++)
     releaseSignal(tSignal[i]);
   theInitState = Initialised; 
+
   DBUG_RETURN(0);
   
 error_handler:
   ndbout << "error_handler" << endl;
   releaseTransactionArrays();
   delete theDictionary;
-  theImpl->m_transporter_facade->close(theNdbBlockNumber, 0);
+  theImpl->close();
   DBUG_RETURN(-1);
 }
 
@@ -167,12 +170,10 @@ Ndb::releaseTransactionArrays()
 }//Ndb::releaseTransactionArrays()
 
 void
-Ndb::executeMessage(void* NdbObject,
-                    NdbApiSignal * aSignal,
-                    LinearSectionPtr ptr[3])
+NdbImpl::trp_deliver_signal(const NdbApiSignal * aSignal,
+                            const LinearSectionPtr ptr[3])
 {
-  Ndb* tNdb = (Ndb*)NdbObject;
-  tNdb->handleReceivedSignal(aSignal, ptr);
+  m_ndb.handleReceivedSignal(aSignal, ptr);
 }
 
 void Ndb::connected(Uint32 ref)
@@ -185,19 +186,12 @@ void Ndb::connected(Uint32 ref)
     assert(theMyRef == numberToRef(theNdbBlockNumber, tmpTheNode));
   }
   
-  TransporterFacade * theFacade =  theImpl->m_transporter_facade;
-  int i, n= 0;
-  for (i = 1; i < MAX_NDB_NODES; i++){
-    if (theFacade->getIsDbNode(i)){
-      theImpl->theDBnodes[n] = i;
-      n++;
-    }
-  }
-  theImpl->theNoOfDBnodes = n;
+  Uint32 cnt =
+    theImpl->m_ndb_cluster_connection.get_db_nodes(theImpl->theDBnodes);
+  theImpl->theNoOfDBnodes = cnt;
   
-  theFirstTransId = ((Uint64)tBlockNo << 52)+
+  theFirstTransId += ((Uint64)tBlockNo << 52)+
     ((Uint64)tmpTheNode << 40);
-  theFirstTransId += theFacade->m_max_trans_id;
   //      assert(0);
   DBUG_PRINT("info",("connected with ref=%x, id=%d, no_db_nodes=%d, first_trans_id: 0x%lx",
 		     theMyRef,
@@ -210,44 +204,6 @@ void Ndb::connected(Uint32 ref)
   theNode= tmpTheNode; // flag that Ndb object is initialized
 }
 
-void Ndb::report_node_connected(Uint32 nodeId)
-{
-  if (theEventBuffer)
-  {
-    // node connected
-    // eventOperations in the ndb object should be notified
-    theEventBuffer->report_node_connected(nodeId);
-  }
-}
-
-void
-Ndb::statusMessage(void* NdbObject, Uint32 a_node, Uint32 _event)
-{
-  DBUG_ENTER("Ndb::statusMessage");
-  NS_Event event = (NS_Event)_event;
-  DBUG_PRINT("info", ("a_node: %u  event: %u",
-                      a_node, _event));
-  Ndb* tNdb = (Ndb*)NdbObject;
-  switch(event){
-  case NS_CONNECTED:
-    // cluster connect, a_node == own reference
-    tNdb->connected(a_node);
-    break;
-  case NS_NODE_ALIVE:
-    tNdb->report_node_connected(a_node);
-    break;
-  case NS_NODE_FAILED:
-    tNdb->report_node_failure(a_node);
-    break;
-  case NS_NODE_NF_COMPLETE:
-    tNdb->report_node_failure_completed(a_node);
-    break;
-  }//if
-  NdbDictInterface::execNodeStatus(&tNdb->theDictionary->m_receiver,
-				   a_node, event);
-  DBUG_VOID_RETURN;
-}
-
 void
 Ndb::report_node_failure(Uint32 node_id)
 {
@@ -341,20 +297,35 @@ Parameters:     aSignal: The signal object.
 Remark:         Send all operations belonging to this connection. 
 *****************************************************************************/
 void	
-Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
+Ndb::handleReceivedSignal(const NdbApiSignal* aSignal,
+			  const LinearSectionPtr ptr[3])
 {
   NdbOperation* tOp;
   NdbIndexOperation* tIndexOp;
   NdbTransaction* tCon;
   int tReturnCode = -1;
   const Uint32* tDataPtr = aSignal->getDataPtr();
-  const Uint32 tWaitState = theImpl->theWaiter.m_state;
+  const Uint32 tWaitState = theImpl->theWaiter.get_state();
   const Uint32 tSignalNumber = aSignal->readSignalNumber();
   const Uint32 tFirstData = *tDataPtr;
   const Uint32 tLen = aSignal->getLength();
+  Uint32 tNewState = tWaitState;
   void * tFirstDataPtr;
-  NdbWaiter *t_waiter;
+  NdbWaiter *t_waiter = &theImpl->theWaiter;
 
+  /* Update cached Min Db node version */
+  theCachedMinDbNodeVersion = theImpl->m_transporter_facade->getMinDbNodeVersion();
+
+  if (likely(NdbImpl::recordGSN(tSignalNumber)))
+  {
+    Uint32 secs = aSignal->m_noOfSections;
+    theImpl->incClientStat(BytesRecvdCount,
+                           ((aSignal->getLength() << 2) +
+                            ((secs > 2)? ptr[2].sz << 2: 0) + 
+                            ((secs > 1)? ptr[1].sz << 2: 0) +
+                            ((secs > 0)? ptr[0].sz << 2: 0)));
+  }
+  
   /*
     In order to support 64 bit processes in the application we need to use
     id's rather than a direct pointer to the object used. It is also a good
@@ -367,14 +338,21 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
   
   switch (tSignalNumber){
   case GSN_TCKEYCONF:
+  case GSN_TCINDXCONF:
     {
-      tFirstDataPtr = int2void(tFirstData);
-      if (tFirstDataPtr == 0) goto InvalidSignal;
-
       const TcKeyConf * const keyConf = (TcKeyConf *)tDataPtr;
+      if (tFirstData != RNIL)
+      {
+        tFirstDataPtr = int2void(tFirstData);
+        if (tFirstDataPtr == 0) goto InvalidSignal;
+        tCon = void2con(tFirstDataPtr);
+      }
+      else
+      {
+        tCon = lookupTransactionFromOperation(keyConf);
+      }
       const BlockReference aTCRef = aSignal->theSendersBlockRef;
 
-      tCon = void2con(tFirstDataPtr);
       if ((tCon->checkMagicNumber() == 0) &&
           (tCon->theSendStatus == NdbTransaction::sendTC_OP)) {
         tReturnCode = tCon->receiveTCKEYCONF(keyConf, tLen);
@@ -383,7 +361,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
         }//if
 
 	if(TcKeyConf::getMarkerFlag(keyConf->confInfo)){
-	  NdbTransaction::sendTC_COMMIT_ACK(theImpl->m_transporter_facade,
+	  NdbTransaction::sendTC_COMMIT_ACK(theImpl,
                                             theCommitAckSignal,
                                             keyConf->transId1, 
                                             keyConf->transId2,
@@ -398,42 +376,76 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
     }
   case GSN_TRANSID_AI:{
     tFirstDataPtr = int2void(tFirstData);
-    NdbReceiver* tRec;
-    if (tFirstDataPtr && (tRec = void2rec(tFirstDataPtr)) && 
-	tRec->checkMagicNumber() && (tCon = tRec->getTransaction()) &&
-	tCon->checkState_TransId(((const TransIdAI*)tDataPtr)->transId)){
-      Uint32 com;
-      if(aSignal->m_noOfSections > 0){
-	com = tRec->execTRANSID_AI(ptr[0].p, ptr[0].sz);
-      } else {
-	com = tRec->execTRANSID_AI(tDataPtr + TransIdAI::HeaderLength, 
-				   tLen - TransIdAI::HeaderLength);
-      }
-
-      if(com == 0)
+    if (tFirstDataPtr){
+      NdbReceiver* const tRec = void2rec(tFirstDataPtr);
+      if(!tRec->checkMagicNumber()){
 	return;
-      
-      switch(tRec->getType()){
-      case NdbReceiver::NDB_OPERATION:
-      case NdbReceiver::NDB_INDEX_OPERATION:
-	if(tCon->OpCompleteSuccess() != -1){
-	  completedTransaction(tCon);
+      }
+      tCon = tRec->getTransaction();
+      if((tCon!=NULL) &&
+	 tCon->checkState_TransId(((const TransIdAI*)tDataPtr)->transId)){
+	Uint32 com;
+	if(aSignal->m_noOfSections > 0){
+	  if(tRec->getType()==NdbReceiver::NDB_QUERY_OPERATION){
+	    com = ((NdbQueryOperationImpl*)(tRec->m_owner))
+              ->execTRANSID_AI(ptr[0].p, ptr[0].sz);
+	  }else{
+	    com = tRec->execTRANSID_AI(ptr[0].p, ptr[0].sz);
+	  }
+	} else {
+	  assert(tRec->getType()!=NdbReceiver::NDB_QUERY_OPERATION);
+	  com = tRec->execTRANSID_AI(tDataPtr + TransIdAI::HeaderLength, 
+				     tLen - TransIdAI::HeaderLength);
+	}
+
+        {
+          tCon->theNdb->theImpl->incClientStat(Ndb::ReadRowCount, 1);
+          if (refToNode(aSignal->theSendersBlockRef) == tCon->theDBnode)
+            tCon->theNdb->theImpl->incClientStat(Ndb::TransLocalReadRowCount,1);
+        }
+
+	if(com == 0)
+	  return;
+
+	switch(tRec->getType()){
+	case NdbReceiver::NDB_OPERATION:
+	case NdbReceiver::NDB_INDEX_OPERATION:
+	  if(tCon->OpCompleteSuccess() != -1){ //More completions pending?
+	    completedTransaction(tCon);
+	  }
+	  return;
+	case NdbReceiver::NDB_SCANRECEIVER:
+	  tCon->theScanningOp->receiver_delivered(tRec);
+          tNewState = (((WaitSignalType) tWaitState) == WAIT_SCAN ? 
+                     (Uint32) NO_WAIT : tWaitState);
+	  break;
+        case NdbReceiver::NDB_QUERY_OPERATION:
+        {
+          // Handled differently whether it is a scan or lookup
+          NdbQueryOperationImpl* tmp = (NdbQueryOperationImpl*)(tRec->m_owner);
+          if (tmp->getQueryDef().isScanQuery()) {
+            tNewState = (((WaitSignalType) tWaitState) == WAIT_SCAN ? 
+                       (Uint32) NO_WAIT : tWaitState);
+            break;
+          } else {
+            if (tCon->OpCompleteSuccess() != -1) { //More completions pending?
+              completedTransaction(tCon);
+            }
+            return;
+          }
+        }
+	default:
+	  goto InvalidSignal;
 	}
-	return;
-      case NdbReceiver::NDB_SCANRECEIVER:
-	tCon->theScanningOp->receiver_delivered(tRec);
-	theImpl->theWaiter.m_state = (((WaitSignalType) tWaitState) == WAIT_SCAN ? 
-				      (Uint32) NO_WAIT : tWaitState);
 	break;
-      default:
-	goto InvalidSignal;
+      } else {
+	/**
+	 * This is ok as transaction can have been aborted before TRANSID_AI
+	 * arrives (if TUP on  other node than TC)
+	 */
+	return;
       }
-      break;
-    } else {
-      /**
-       * This is ok as transaction can have been aborted before TRANSID_AI
-       * arrives (if TUP on  other node than TC)
-       */
+    } else{ // if((tFirstDataPtr)
       return;
     }
   }
@@ -463,7 +475,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
 #endif
       }
       if(tFirstData & 1){
-	NdbTransaction::sendTC_COMMIT_ACK(theImpl->m_transporter_facade,
+	NdbTransaction::sendTC_COMMIT_ACK(theImpl,
                                           theCommitAckSignal,
                                           failConf->transId1, 
                                           failConf->transId2,
@@ -501,20 +513,35 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       tFirstDataPtr = int2void(tFirstData);
       if (tFirstDataPtr == 0) goto InvalidSignal;
 
-      tOp = void2rec_op(tFirstDataPtr);
-      if (tOp->checkMagicNumber() == 0) {
-	tCon = tOp->theNdbCon;
-	if (tCon != NULL) {
-	  if (tCon->theSendStatus == NdbTransaction::sendTC_OP) {
-	    tReturnCode = tOp->receiveTCKEYREF(aSignal);
-	    if (tReturnCode != -1) {
-	      completedTransaction(tCon);
-	      return;
-	    }//if
-	    break;
-	  }//if
-	}//if
-      } //if
+      const NdbReceiver* const receiver = void2rec(tFirstDataPtr);
+      if(!receiver->checkMagicNumber()){
+        goto InvalidSignal; 
+      }
+      tCon = receiver->getTransaction();
+      if (tCon != NULL) {
+        if (tCon->theSendStatus == NdbTransaction::sendTC_OP) {
+          if (receiver->getType()==NdbReceiver::NDB_QUERY_OPERATION) {
+            NdbQueryOperationImpl* tmp =
+              (NdbQueryOperationImpl*)(receiver->m_owner);
+            if (tmp->execTCKEYREF(aSignal) &&
+               tCon->OpCompleteFailure() != -1) {
+              completedTransaction(tCon);
+              return;
+            }
+          } else {
+            tOp = void2rec_op(tFirstDataPtr);
+            /* NB! NdbOperation::checkMagicNumber() returns 0 if it *is* 
+             * an NdbOperation.*/
+            assert(tOp->checkMagicNumber()==0); 
+            tReturnCode = tOp->receiveTCKEYREF(aSignal);
+            if (tReturnCode != -1) {
+              completedTransaction(tCon);
+              return;
+            }//if
+          }//if
+          break;
+        }//if
+      }//if (tCon != NULL)
       goto InvalidSignal;
       return;
     } 
@@ -535,7 +562,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
 	}//if
 
 	if(tFirstData & 1){
-	  NdbTransaction::sendTC_COMMIT_ACK(theImpl->m_transporter_facade,
+	  NdbTransaction::sendTC_COMMIT_ACK(theImpl,
                                             theCommitAckSignal,
                                             commitConf->transId1, 
                                             commitConf->transId2,
@@ -620,7 +647,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       }//if
       tReturnCode = tCon->receiveTCSEIZECONF(aSignal);
       if (tReturnCode != -1) {
-	theImpl->theWaiter.m_state = NO_WAIT;
+        tNewState = NO_WAIT;
       } else {
 	goto InvalidSignal;
       }//if
@@ -640,7 +667,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       }//if
       tReturnCode = tCon->receiveTCSEIZEREF(aSignal);
       if (tReturnCode != -1) {
-	theImpl->theWaiter.m_state = NO_WAIT;
+        tNewState = NO_WAIT;
       } else {
         return;
       }//if
@@ -660,7 +687,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       }//if
       tReturnCode = tCon->receiveTCRELEASECONF(aSignal);
       if (tReturnCode != -1) {
-	theImpl->theWaiter.m_state = NO_WAIT;
+        tNewState = NO_WAIT;
       }//if
       break;
     } 
@@ -678,7 +705,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       }//if
       tReturnCode = tCon->receiveTCRELEASEREF(aSignal);
       if (tReturnCode != -1) {
-	theImpl->theWaiter.m_state = NO_WAIT;
+        tNewState = NO_WAIT;
       }//if
       break;
     }
@@ -695,6 +722,8 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
   case GSN_CREATE_INDX_REF:
   case GSN_DROP_INDX_CONF:
   case GSN_DROP_INDX_REF:
+  case GSN_INDEX_STAT_CONF:
+  case GSN_INDEX_STAT_REF:
   case GSN_CREATE_EVNT_CONF:
   case GSN_CREATE_EVNT_REF:
   case GSN_DROP_EVNT_CONF:
@@ -708,8 +737,15 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
   case GSN_DROP_FILE_CONF:
   case GSN_DROP_FILEGROUP_REF:
   case GSN_DROP_FILEGROUP_CONF:
+  case GSN_SCHEMA_TRANS_BEGIN_CONF:
+  case GSN_SCHEMA_TRANS_BEGIN_REF:
+  case GSN_SCHEMA_TRANS_END_CONF:
+  case GSN_SCHEMA_TRANS_END_REF:
+  case GSN_SCHEMA_TRANS_END_REP:
   case GSN_WAIT_GCP_CONF:
   case GSN_WAIT_GCP_REF:
+  case GSN_CREATE_HASH_MAP_REF:
+  case GSN_CREATE_HASH_MAP_CONF:
     NdbDictInterface::execSignal(&theDictionary->m_receiver,
 				 aSignal, ptr);
     return;
@@ -751,9 +787,15 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
         !op->execSUB_TABLE_DATA(aSignal, ptr))
       return;
     
-    for (int i= aSignal->m_noOfSections;i < 3; i++) {
-      ptr[i].p = NULL;
-      ptr[i].sz = 0;
+    LinearSectionPtr copy[3];
+    for (int i = 0; i<aSignal->m_noOfSections; i++)
+    {
+      copy[i] = ptr[i];
+    }
+    for (int i = aSignal->m_noOfSections; i < 3; i++)
+    {
+      copy[i].p = NULL;
+      copy[i].sz = 0;
     }
     DBUG_PRINT("info",("oid=senderData: %d, gci{hi/lo}: %d/%d, operation: %d, "
 		       "tableId: %d",
@@ -761,24 +803,9 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
 		       SubTableData::getOperation(sdata->requestInfo),
 		       sdata->tableId));
 
-    theEventBuffer->insertDataL(op, sdata, tLen, ptr);
+    theEventBuffer->insertDataL(op, sdata, tLen, copy);
     return;
   }
-  case GSN_DIHNDBTAMPER:
-    {
-      tFirstDataPtr = int2void(tFirstData);
-      if (tFirstDataPtr == 0) goto InvalidSignal;
-      
-      if (tWaitState != WAIT_NDB_TAMPER)
-	return;
-      tCon = void2con(tFirstDataPtr);
-      if (tCon->checkMagicNumber() != 0)
-	return;
-      tReturnCode = tCon->receiveDIHNDBTAMPER(aSignal);
-      if (tReturnCode != -1)
-	theImpl->theWaiter.m_state = NO_WAIT;
-      break;
-    }
   case GSN_SCAN_TABCONF:
     {
       tFirstDataPtr = int2void(tFirstData);
@@ -798,7 +825,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
 				      tLen - ScanTabConf::SignalLength);
 	}
 	if (tReturnCode != -1 && tWaitState == WAIT_SCAN)
-	  theImpl->theWaiter.m_state = NO_WAIT;
+          tNewState = NO_WAIT;
 	break;
       } else {
 	goto InvalidSignal;
@@ -817,7 +844,7 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       if (tCon->checkMagicNumber() == 0){
 	tReturnCode = tCon->receiveSCAN_TABREF(aSignal);
 	if (tReturnCode != -1 && tWaitState == WAIT_SCAN){
-	  theImpl->theWaiter.m_state = NO_WAIT;
+          tNewState = NO_WAIT;
 	}
 	break;
       }
@@ -842,8 +869,8 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       switch(com){
       case 1:
 	tCon->theScanningOp->receiver_delivered(tRec);
-	theImpl->theWaiter.m_state = (((WaitSignalType) tWaitState) == WAIT_SCAN ? 
-			      (Uint32) NO_WAIT : tWaitState);
+        tNewState = (((WaitSignalType) tWaitState) == WAIT_SCAN ? 
+                     (Uint32) NO_WAIT : tWaitState);
 	break;
       case 0:
 	break;
@@ -859,30 +886,6 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       return;
     }
   }
-  case GSN_TCINDXCONF:{
-    tFirstDataPtr = int2void(tFirstData);
-    if (tFirstDataPtr == 0) goto InvalidSignal;
-
-    const TcIndxConf * const indxConf = (TcIndxConf *)tDataPtr;
-    const BlockReference aTCRef = aSignal->theSendersBlockRef;
-    tCon = void2con(tFirstDataPtr);
-    if ((tCon->checkMagicNumber() == 0) &&
-	(tCon->theSendStatus == NdbTransaction::sendTC_OP)) {
-      tReturnCode = tCon->receiveTCINDXCONF(indxConf, tLen);
-      if (tReturnCode != -1) { 
-	completedTransaction(tCon);
-      }//if
-    }//if
-    
-    if(TcIndxConf::getMarkerFlag(indxConf->confInfo)){
-      NdbTransaction::sendTC_COMMIT_ACK(theImpl->m_transporter_facade,
-                                        theCommitAckSignal,
-                                        indxConf->transId1, 
-                                        indxConf->transId2,
-                                        aTCRef);
-    }
-    return;
-  }
   case GSN_TCINDXREF:{
     tFirstDataPtr = int2void(tFirstData);
     if (tFirstDataPtr == 0) goto InvalidSignal;
@@ -903,15 +906,47 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
     goto InvalidSignal;
     return;
   } 
+  case GSN_API_REGCONF:
+  case GSN_CONNECT_REP:
+    return; // Ignore
+  case GSN_NODE_FAILREP:
+  {
+    const NodeFailRep *rep = CAST_CONSTPTR(NodeFailRep,
+                                           aSignal->getDataPtr());
+    for (Uint32 i = NdbNodeBitmask::find_first(rep->theNodes);
+         i != NdbNodeBitmask::NotFound;
+         i = NdbNodeBitmask::find_next(rep->theNodes, i + 1))
+    {
+      report_node_failure(i);
+    }
+
+    NdbDictInterface::execSignal(&theDictionary->m_receiver, aSignal, ptr);
+    break;
+  }
+  case GSN_NF_COMPLETEREP:
+  {
+    const NFCompleteRep *rep = CAST_CONSTPTR(NFCompleteRep,
+                                             aSignal->getDataPtr());
+    report_node_failure_completed(rep->failedNodeId);
+    break;
+  }
   case GSN_TAKE_OVERTCCONF:
     abortTransactionsAfterNodeFailure(tFirstData); // theData[0]
     break;
+  case GSN_ALLOC_NODEID_CONF:
+  {
+    const AllocNodeIdConf *rep = CAST_CONSTPTR(AllocNodeIdConf,
+                                               aSignal->getDataPtr());
+    Uint32 nodeId = rep->nodeId;
+    connected(numberToRef(theNdbBlockNumber, nodeId));
+    break;
+  }
   default:
+    tFirstDataPtr = NULL;
     goto InvalidSignal;
   }//swich
 
-  t_waiter= &theImpl->theWaiter;
-  if (t_waiter->get_state() == NO_WAIT && tWaitState != NO_WAIT)
+  if (tNewState != tWaitState)
   {
     /*
       If our waiter object is the owner of the "poll rights", then we
@@ -924,20 +959,13 @@ Ndb::handleReceivedSignal(NdbApiSignal* aSignal, LinearSectionPtr ptr[3])
       its conditional wait. This will wake up this thread so that it
       can continue its work.
     */
-    TransporterFacade *tp= theImpl->m_transporter_facade;
-    if (tp->get_poll_owner() != t_waiter)
-    {
-      /*
-         Wake up the thread waiting for response and remove it from queue
-         of objects waiting for receive completion
-      */
-      tp->remove_from_cond_wait_queue(t_waiter);
-      t_waiter->cond_signal();
-    }
-  }//if
+    t_waiter->signal(tNewState);
+  }
+
   return;
 
- InvalidSignal:
+
+InvalidSignal:
 #ifdef VM_TRACE
   ndbout_c("Ndbif: Error Ndb::handleReceivedSignal "
 	   "(tFirstDataPtr=%p, GSN=%d, theImpl->theWaiter.m_state=%d)"
@@ -986,24 +1014,24 @@ Ndb::completedTransaction(NdbTransaction* aCon)
     theNoOfSentTransactions = tNoSentTransactions - 1;
     aCon->theListState = NdbTransaction::InCompletedList;
     aCon->handleExecuteCompletion();
-    if ((theMinNoOfEventsToWakeUp != 0) &&
-        (theNoOfCompletedTransactions >= theMinNoOfEventsToWakeUp)) {
-      theMinNoOfEventsToWakeUp = 0;
-      TransporterFacade *tp = theImpl->m_transporter_facade;
-      NdbWaiter *t_waiter= &theImpl->theWaiter;
-      if (tp->get_poll_owner() != t_waiter) {
-        /*
-          When we come here, this is executed by the thread owning the "poll
-          rights". This thread is not where our waiter object belongs.
-          Thus we wake up the thread owning this waiter object but first
-          we must remove it from the conditional wait queue so that we
-          don't assign it as poll owner later on.
-        */
-        tp->remove_from_cond_wait_queue(t_waiter);
-        t_waiter->cond_signal();
+
+    if (theImpl->wakeHandler == 0)
+    {
+      if ((theMinNoOfEventsToWakeUp != 0) &&
+          (theNoOfCompletedTransactions >= theMinNoOfEventsToWakeUp))
+      {
+        theMinNoOfEventsToWakeUp = 0;
+        theImpl->theWaiter.signal(NO_WAIT);
+        return;
       }
-      return;
-    }//if
+    }
+    else
+    {
+      /**
+       * This is for multi-wait handling
+       */
+      theImpl->wakeHandler->notifyTransactionCompleted(this);
+    }
   } else {
     ndbout << "theNoOfSentTransactions = " << (int) theNoOfSentTransactions;
     ndbout << " theListState = " << (int) aCon->theListState;
@@ -1071,7 +1099,7 @@ Ndb::pollCompleted(NdbTransaction** aCopyArray)
 void
 Ndb::check_send_timeout()
 {
-  Uint32 timeout = theImpl->m_transporter_facade->m_waitfor_timeout;
+  Uint32 timeout = theImpl->get_ndbapi_config_parameters().m_waitfor_timeout;
   NDB_TICKS current_time = NdbTick_CurrentMillisecond();
   assert(current_time >= the_last_check_time);
   if (current_time - the_last_check_time > 1000) {
@@ -1083,7 +1111,7 @@ Ndb::check_send_timeout()
       {
 #ifdef VM_TRACE
         a_con->printState();
-	Uint32 t1 = a_con->theTransactionId;
+	Uint32 t1 = (Uint32) a_con->theTransactionId;
 	Uint32 t2 = a_con->theTransactionId >> 32;
 	ndbout_c("4012 [%.8x %.8x]", t1, t2);
 	//abort();
@@ -1161,14 +1189,14 @@ Ndb::sendPrepTrans(int forceSend)
      and we keep a small space for messages like that.
   */
   Uint32 i;
-  TransporterFacade* tp = theImpl->m_transporter_facade;
+  theCachedMinDbNodeVersion = theImpl->m_transporter_facade->getMinDbNodeVersion();
   Uint32 no_of_prep_trans = theNoOfPreparedTransactions;
   for (i = 0; i < no_of_prep_trans; i++) {
     NdbTransaction * a_con = thePreparedTransactionsArray[i];
     thePreparedTransactionsArray[i] = NULL;
     Uint32 node_id = a_con->getConnectedNodeId();
-    if ((tp->getNodeSequence(node_id) == a_con->theNodeSequence) &&
-        (tp->get_node_alive(node_id) || tp->get_node_stopping(node_id)))
+    if ((theImpl->getNodeSequence(node_id) == a_con->theNodeSequence) &&
+        (theImpl->get_node_alive(node_id) || theImpl->get_node_stopping(node_id)))
     {
       /*
       We will send if
@@ -1178,7 +1206,7 @@ Ndb::sendPrepTrans(int forceSend)
       of all transactions and commits and thus we allow aborts and
       commits to continue but not normal operations.
       */
-      if (tp->check_send_size(node_id, a_con->get_send_size())) {
+      if (theImpl->check_send_size(node_id, a_con->get_send_size())) {
         if (a_con->doSend() == 0) {
           NDB_TICKS current_time = NdbTick_CurrentMillisecond();
           a_con->theStartTransTime = current_time;
@@ -1231,11 +1259,13 @@ Ndb::sendPrepTrans(int forceSend)
     insert_completed_list(a_con);
   }//for
   theNoOfPreparedTransactions = 0;
-  if (forceSend == 0) {
-     tp->checkForceSend(theNdbBlockNumber);
-  } else if (forceSend == 1) {
-     tp->forceSend(theNdbBlockNumber);
-  }//if
+  int did_send = theImpl->do_forceSend(forceSend);
+  if(forceSend) {
+    theImpl->incClientStat(Ndb::ForcedSendsCount, 1);
+  }
+  else {
+    theImpl->incClientStat(did_send ? Ndb::UnforcedSendsCount : Ndb::DeferredSendsCount, 1);
+  }
   return;
 }//Ndb::sendPrepTrans()
 
@@ -1251,9 +1281,11 @@ Ndb::waitCompletedTransactions(int aMilliSecondsToWait,
 			       int noOfEventsToWaitFor,
                                PollGuard *poll_guard)
 {
-  theImpl->theWaiter.m_state = NO_WAIT; 
+  theImpl->theWaiter.set_node(0);
+  theImpl->theWaiter.set_state(WAIT_TRANS);
+
   /**
-   * theImpl->theWaiter.m_state = NO_WAIT; 
+   * theImpl->theWaiter.set_node(0)
    * To ensure no messup with synchronous node fail handling
    * (see ReportFailure)
    */
@@ -1262,6 +1294,7 @@ Ndb::waitCompletedTransactions(int aMilliSecondsToWait,
   NDB_TICKS maxTime = currTime + (NDB_TICKS)waitTime;
   theMinNoOfEventsToWakeUp = noOfEventsToWaitFor;
   const int maxsleep = aMilliSecondsToWait > 10 ? 10 : aMilliSecondsToWait;
+  theImpl->incClientStat(Ndb::WaitExecCompleteCount, 1);
   do {
     poll_guard->wait_for_input(maxsleep);
     if (theNoOfCompletedTransactions >= (Uint32)noOfEventsToWaitFor) {
@@ -1272,10 +1305,6 @@ Ndb::waitCompletedTransactions(int aMilliSecondsToWait,
   } while (waitTime > 0);
 }//Ndb::waitCompletedTransactions()
 
-void Ndb::cond_signal()
-{
-  NdbCondition_Signal(theImpl->theWaiter.m_condition);
-}
 /*****************************************************************************
 void sendPreparedTransactions(int forceSend = 0);
 
@@ -1286,9 +1315,9 @@ Remark:   First send all prepared operations and then check if there are any
 void	
 Ndb::sendPreparedTransactions(int forceSend)
 {
-  theImpl->m_transporter_facade->lock_mutex();
+  theImpl->lock();
   sendPrepTrans(forceSend);
-  theImpl->m_transporter_facade->unlock_mutex();
+  theImpl->unlock();
   return;
 }//Ndb::sendPreparedTransactions()
 
@@ -1309,8 +1338,7 @@ Ndb::sendPollNdb(int aMillisecondNumber, int minNoOfEventsToWakeup, int forceSen
     in all places where the object is out of context due to a return,
     break, continue or simply end of statement block
   */
-  PollGuard pg(theImpl->m_transporter_facade, &theImpl->theWaiter,
-               theNdbBlockNumber);
+  PollGuard pg(* theImpl);
   sendPrepTrans(forceSend);
   return poll_trans(aMillisecondNumber, minNoOfEventsToWakeup, &pg);
 }
@@ -1354,8 +1382,7 @@ Ndb::pollNdb(int aMillisecondNumber, int minNoOfEventsToWakeup)
     in all places where the object is out of context due to a return,
     break, continue or simply end of statement block
   */
-  PollGuard pg(theImpl->m_transporter_facade, &theImpl->theWaiter,
-               theNdbBlockNumber);
+  PollGuard pg(* theImpl);
   return poll_trans(aMillisecondNumber, minNoOfEventsToWakeup, &pg);
 }
 
@@ -1378,7 +1405,6 @@ Ndb::sendRecSignal(Uint16 node_id,
 
   int return_code;
   Uint32 read_conn_seq;
-  TransporterFacade* tp = theImpl->m_transporter_facade;
   Uint32 send_size = 1; // Always sends one signal only
   // Protected area
   /*
@@ -1387,15 +1413,16 @@ Ndb::sendRecSignal(Uint16 node_id,
     in all places where the object is out of context due to a return,
     break, continue or simply end of statement block
   */
-  PollGuard poll_guard(tp,&theImpl->theWaiter,theNdbBlockNumber);
-  read_conn_seq= tp->getNodeSequence(node_id);
+  theImpl->incClientStat(WaitMetaRequestCount, 1);
+  PollGuard poll_guard(* theImpl);
+  read_conn_seq= theImpl->getNodeSequence(node_id);
   if (ret_conn_seq)
     *ret_conn_seq= read_conn_seq;
-  if ((tp->get_node_alive(node_id)) &&
+  if ((theImpl->get_node_alive(node_id)) &&
       ((read_conn_seq == conn_seq) ||
        (conn_seq == 0))) {
-    if (tp->check_send_size(node_id, send_size)) {
-      return_code = tp->sendSignal(aSignal, node_id);
+    if (theImpl->check_send_size(node_id, send_size)) {
+      return_code = theImpl->sendSignal(aSignal, node_id);
       if (return_code != -1) {
         return poll_guard.wait_n_unlock(WAITFOR_RESPONSE_TIMEOUT,node_id,
                                          aWaitState, false);
@@ -1406,7 +1433,7 @@ Ndb::sendRecSignal(Uint16 node_id,
       return_code = -4;
     }//if
   } else {
-    if ((tp->get_node_stopping(node_id)) &&
+    if ((theImpl->get_node_stopping(node_id)) &&
         ((read_conn_seq == conn_seq) ||
          (conn_seq == 0))) {
       return_code = -5;
@@ -1419,7 +1446,7 @@ Ndb::sendRecSignal(Uint16 node_id,
 }//Ndb::sendRecSignal()
 
 void
-NdbTransaction::sendTC_COMMIT_ACK(TransporterFacade *tp,
+NdbTransaction::sendTC_COMMIT_ACK(NdbImpl * impl,
                                   NdbApiSignal * aSignal,
                                   Uint32 transId1, Uint32 transId2, 
                                   Uint32 aTCRef){
@@ -1430,14 +1457,14 @@ NdbTransaction::sendTC_COMMIT_ACK(TransporterFacade *tp,
 	   refToNode(aTCRef));
 #endif  
   aSignal->theTrace                = TestOrd::TraceAPI;
-  aSignal->theReceiversBlockNumber = DBTC;
+  aSignal->theReceiversBlockNumber = refToBlock(aTCRef);
   aSignal->theVerId_signalNumber   = GSN_TC_COMMIT_ACK;
   aSignal->theLength               = 2;
 
   Uint32 * dataPtr = aSignal->getDataPtrSend();
   dataPtr[0] = transId1;
   dataPtr[1] = transId2;
-  tp->sendSignalUnCond(aSignal, refToNode(aTCRef), 1);
+  impl->safe_sendSignal(aSignal, refToNode(aTCRef));
 }
 
 int
@@ -1445,27 +1472,28 @@ NdbImpl::send_event_report(bool has_lock,
                            Uint32 *data, Uint32 length)
 {
   NdbApiSignal aSignal(m_ndb.theMyRef);
-  TransporterFacade *tp = m_transporter_facade;
   aSignal.theTrace                = TestOrd::TraceAPI;
   aSignal.theReceiversBlockNumber = CMVMI;
   aSignal.theVerId_signalNumber   = GSN_EVENT_REP;
   aSignal.theLength               = length;
   memcpy((char *)aSignal.getDataPtrSend(), (char *)data, length*4);
 
-
   int ret = 0;
   if (!has_lock)
   {
-    tp->lock_mutex();
+    lock();
   }
   Uint32 tNode;
   Ndb_cluster_connection_node_iter node_iter;
   m_ndb_cluster_connection.init_get_next_node(node_iter);
   while ((tNode= m_ndb_cluster_connection.get_next_node(node_iter)))
   {
-    if(tp->get_node_alive(tNode))
+    if(get_node_alive(tNode))
     {
-      tp->sendSignal(&aSignal, tNode);
+      if (has_lock)
+        safe_sendSignal(&aSignal, tNode);
+      else
+        raw_sendSignal(&aSignal, tNode);
       goto done;
     }
   }
@@ -1474,7 +1502,24 @@ NdbImpl::send_event_report(bool has_lock,
 done:
   if (!has_lock)
   {
-    tp->unlock_mutex();
+    unlock();
   }
   return ret;
 }
+
+NdbTransaction*
+Ndb::lookupTransactionFromOperation(const TcKeyConf * conf)
+{
+  assert(TcKeyConf::getNoOfOperations(conf->confInfo) > 0);
+  Uint32 opPtr = conf->operations[0].apiOperationPtr;
+  void * voidptr = int2void(opPtr);
+  if (voidptr)
+  {
+    NdbReceiver* rec = void2rec(voidptr);
+    if (rec)
+    {
+      return rec->getTransaction();
+    }
+  }
+  return 0;
+}
