@@ -39,6 +39,8 @@ static int   minEventSendPoll;
 static int   forceSendPoll;
 static bool  useNdbRecord;
 static bool  useCombUpd;
+int          subscriberCount;
+static bool  robustMode;
 
 static ThreadData *data;
 static Ndb_cluster_connection *g_cluster_connection= 0;
@@ -60,7 +62,7 @@ static void usage(const char *prog)
 
    ndbout_c(
            "Usage: %s [-proc <num>] [-warm <num>] [-time <num>] [ -p <num>]" 
-	   "[-t <num> ] [ -e <num> ] [ -f <num>] [ -ndbrecord ]\n"
+	   "[-t <num> ] [ -e <num> ] [ -f <num>] [ -ndbrecord ] [ -s <num>]\n"
            "  -proc <num>    Specifies that <num> is the number of\n"
            "                 threads. The default is 1.\n"
            "  -time <num>    Specifies that the test will run for <num> sec.\n"
@@ -78,8 +80,10 @@ static void usage(const char *prog)
            "  -ndbrecord     Use NdbRecord Api.\n"
            "                 Default is to use old Api\n"
            "  -combupdread   Use update pre-read operation where possible\n"
-           "                 Default is to use separate read+update ops\n",
-           progname);
+           "                 Default is to use separate read+update ops\n"
+           "  -s <num>       Number of subscribers to operate on, default is %u.\n"
+           "  -r             Whether to be robust to key errors\n",
+           progname, NO_OF_SUBSCRIBERS);
 }
 
 static
@@ -97,6 +101,8 @@ parse_args(int argc, const char **argv)
    forceSendPoll    = 0;
    useNdbRecord     = false;
    useCombUpd       = false;
+   subscriberCount  = NO_OF_SUBSCRIBERS;
+   robustMode       = false;
 
    i = 1;
    while (i < argc){
@@ -176,6 +182,20 @@ parse_args(int argc, const char **argv)
        useCombUpd= true;
        i++;
      }
+     else if (strcmp("-s", argv[i]) == 0) {
+       if (i + 1 >= argc) {
+         return 1;
+       }
+       if (sscanf(argv[i+1], "%u", &subscriberCount) == -1) {
+         ndbout_c("-s flag requires a positive argument.");
+         return 1;
+       }
+       i+=2;
+     }
+     else if (strcmp("-r", argv[i]) == 0) {
+       robustMode= true;
+       i++;
+     }
      else {
        return 1;
      }
@@ -543,11 +563,12 @@ NDB_COMMAND(DbAsyncGenerator, "DbAsyncGenerator",
       data[tid].testSeconds           = numSeconds;
       data[tid].coolDownSeconds       = numWarmSeconds;
       data[tid].randomSeed            = 
-	NdbTick_CurrentMillisecond()+i+j;
+	(unsigned long)(NdbTick_CurrentMillisecond()+i+j);
       data[tid].changedTime           = 0;
       data[tid].runState              = Runnable;
       data[tid].ndbRecordSharedData   = ndbRecordSharedDataPtr;
       data[tid].useCombinedUpdate     = useCombUpd;
+      data[tid].robustMode            = robustMode;
     }
     sprintf(threadName, "AsyncThread[%d]", i);
     pThread = NdbThread_Create(threadRoutine, 
@@ -714,7 +735,7 @@ void showTime()
   now = ::time((time_t*)NULL);
   tm_now = ::gmtime(&now);
 
-  ::snprintf(buf, 128,
+  BaseString::snprintf(buf, 128,
 	     "%d-%.2d-%.2d %.2d:%.2d:%.2d", 
 	     tm_now->tm_year + 1900, 
 	     tm_now->tm_mon, 
