@@ -1,6 +1,5 @@
 /*
-   Copyright (c) 2003-2006, 2008 MySQL AB, 2009 Sun Microsystems, Inc.
-   Use is subject to license terms.
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,6 +21,7 @@
 #include <UtilTransactions.hpp>
 #include <NdbBackup.hpp>
 
+int runDropTable(NDBT_Context* ctx, NDBT_Step* step);
 
 #define CHECK(b) if (!(b)) { \
   g_err << "ERR: "<< step->getName() \
@@ -29,10 +29,13 @@
   result = NDBT_FAILED; \
   continue; } 
 
+char tabname[1000];
+
 int
 clearOldBackups(NDBT_Context* ctx, NDBT_Step* step)
 {
-  NdbBackup backup(GETNDB(step)->getNodeId());
+  strcpy(tabname, ctx->getTab()->getName());
+  NdbBackup backup;
   backup.clearOldBackups();
   return NDBT_OK;
 }
@@ -67,7 +70,7 @@ int setSlave(NDBT_Context* ctx, NDBT_Step* step){
 }
 
 int runAbort(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
 
   NdbRestarter restarter;
 
@@ -101,7 +104,7 @@ int runAbort(NDBT_Context* ctx, NDBT_Step* step){
 }
 
 int runFail(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
 
   NdbRestarter restarter;
 
@@ -135,7 +138,7 @@ int runFail(NDBT_Context* ctx, NDBT_Step* step){
 }
 
 int runBackupOne(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   unsigned backupId = 0;
 
   if (backup.start(backupId) == -1){
@@ -148,7 +151,7 @@ int runBackupOne(NDBT_Context* ctx, NDBT_Step* step){
 }
 
 int runBackupRandom(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   unsigned backupId = rand() % (MAX_BACKUPS);
 
   if (backup.start(backupId) == -1){
@@ -162,7 +165,7 @@ int runBackupRandom(NDBT_Context* ctx, NDBT_Step* step){
 
 int
 runBackupLoop(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   
   int loops = ctx->getNumLoops();
   while(!ctx->isTestStopped() && loops--)
@@ -217,10 +220,8 @@ runDDL(NDBT_Context* ctx, NDBT_Step* step){
 int runDropTablesRestart(NDBT_Context* ctx, NDBT_Step* step){
   NdbRestarter restarter;
 
-  Ndb* pNdb = GETNDB(step);
-
-  const NdbDictionary::Table *tab = ctx->getTab();
-  pNdb->getDictionary()->dropTable(tab->getName());
+  if (runDropTable(ctx, step) != 0)
+    return NDBT_FAILED;
 
   if (restarter.restartAll(false) != 0)
     return NDBT_FAILED;
@@ -232,7 +233,7 @@ int runDropTablesRestart(NDBT_Context* ctx, NDBT_Step* step){
 }
 
 int runRestoreOne(NDBT_Context* ctx, NDBT_Step* step){
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   unsigned backupId = ctx->getProperty("BackupId"); 
 
   ndbout << "Restoring backup " << backupId << endl;
@@ -241,7 +242,17 @@ int runRestoreOne(NDBT_Context* ctx, NDBT_Step* step){
     return NDBT_FAILED;
   }
 
-  return NDBT_OK;
+  Ndb* pNdb = GETNDB(step);
+  pNdb->getDictionary()->invalidateTable(tabname);
+  const NdbDictionary::Table* tab = pNdb->getDictionary()->getTable(tabname);
+  
+  if (tab)
+  {
+    ctx->setTab(tab);
+    return NDBT_OK;
+  }
+
+  return NDBT_FAILED;
 }
 
 int runVerifyOne(NDBT_Context* ctx, NDBT_Step* step){
@@ -250,8 +261,7 @@ int runVerifyOne(NDBT_Context* ctx, NDBT_Step* step){
   int result = NDBT_OK;
   int count = 0;
 
-  const NdbDictionary::Table* tab = 
-    GETNDB(step)->getDictionary()->getTable(ctx->getTab()->getName());
+  const NdbDictionary::Table* tab = ctx->getTab();
   if(tab == 0)
     return NDBT_FAILED;
   
@@ -287,8 +297,12 @@ int runClearTable(NDBT_Context* ctx, NDBT_Step* step){
   return NDBT_OK;
 }
 
-int runDropTable(NDBT_Context* ctx, NDBT_Step* step){
-  GETNDB(step)->getDictionary()->dropTable(ctx->getTab()->getName());
+int runDropTable(NDBT_Context* ctx, NDBT_Step* step)
+{
+  
+  const NdbDictionary::Table *tab = ctx->getTab();
+  GETNDB(step)->getDictionary()->dropTable(tab->getName());
+  
   return NDBT_OK;
 }
 
@@ -296,8 +310,8 @@ int runDropTable(NDBT_Context* ctx, NDBT_Step* step){
 
 int runCreateBank(NDBT_Context* ctx, NDBT_Step* step){
   Bank bank(ctx->m_cluster_connection);
-  int overWriteExisting = true;
-  if (bank.createAndLoadBank(overWriteExisting, 10) != NDBT_OK)
+  bool overWriteExisting = true;
+  if (bank.createAndLoadBank(overWriteExisting) != NDBT_OK)
     return NDBT_FAILED;
   return NDBT_OK;
 }
@@ -365,7 +379,7 @@ int runBackupBank(NDBT_Context* ctx, NDBT_Step* step){
   int l = 0;
   int maxSleep = 30; // Max seconds between each backup
   Ndb* pNdb = GETNDB(step);
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   unsigned minBackupId = ~0;
   unsigned maxBackupId = 0;
   unsigned backupId = 0;
@@ -411,7 +425,7 @@ int runBackupBank(NDBT_Context* ctx, NDBT_Step* step){
 
 int runRestoreBankAndVerify(NDBT_Context* ctx, NDBT_Step* step){
   NdbRestarter restarter;
-  NdbBackup backup(GETNDB(step)->getNodeId()+1);
+  NdbBackup backup;
   unsigned minBackupId = ctx->getProperty("MinBackupId");
   unsigned maxBackupId = ctx->getProperty("MaxBackupId");
   unsigned backupId = minBackupId;
@@ -427,7 +441,6 @@ int runRestoreBankAndVerify(NDBT_Context* ctx, NDBT_Step* step){
     // TEMPORARY FIX
     // To erase all tables from cache(s)
     // To be removed, maybe replaced by ndb.invalidate();
-    runDropTable(ctx,step);
     {
       Bank bank(ctx->m_cluster_connection);
       
@@ -445,6 +458,10 @@ int runRestoreBankAndVerify(NDBT_Context* ctx, NDBT_Step* step){
     if (restarter.waitClusterStarted() != 0)
       return NDBT_FAILED;
 
+    ndbout << "Dropping " << tabname << endl;
+    NdbDictionary::Dictionary* pDict = GETNDB(step)->getDictionary();
+    pDict->dropTable(tabname);
+
     ndbout << "Restoring backup " << backupId << endl;
     if (backup.restore(backupId) == -1){
       return NDBT_FAILED;
@@ -481,6 +498,180 @@ int runRestoreBankAndVerify(NDBT_Context* ctx, NDBT_Step* step){
   
   return result;
 }
+int runBackupUndoWaitStarted(NDBT_Context* ctx, NDBT_Step* step){
+  NdbBackup backup;
+  unsigned backupId = 0;
+  int undoError = 10041;
+  NdbRestarter restarter;
+
+  if(restarter.waitClusterStarted(60)){
+    g_err << "waitClusterStarted failed"<< endl;
+    return NDBT_FAILED;
+  }
+
+  if (restarter.insertErrorInAllNodes(undoError) != 0) {
+    g_err << "Error insert failed" << endl;
+    return NDBT_FAILED;
+  } 
+  // start backup wait started
+  if (backup.start(backupId, 1, 0, 1) == -1){
+    return NDBT_FAILED;
+  }
+  ndbout << "Started backup " << backupId << endl;
+  ctx->setProperty("BackupId", backupId);
+
+  return NDBT_OK;
+}
+int runChangeUndoDataDuringBackup(NDBT_Context* ctx, NDBT_Step* step){
+  Ndb* pNdb= GETNDB(step);
+
+  int records = ctx->getNumRecords();
+  int num = 5;
+  if (records - 5 < 0)
+    num = 1;
+  
+  HugoTransactions hugoTrans(*ctx->getTab());
+
+  //update all rows
+  if(hugoTrans.pkUpdateRecords(pNdb, records) != 0) {
+    g_err << "Can't update all the records" << endl;
+    return NDBT_FAILED;
+  }
+
+  //delete first 10 rows
+  if(hugoTrans.pkDelRecords(pNdb, num*2) != 0) {
+    g_err << "Can't delete first 5 rows" << endl;
+    return NDBT_FAILED;
+  }
+
+  //add 5 new rows at the first(0 ~ 4)
+  NdbTransaction *pTransaction= pNdb->startTransaction();
+  if (pTransaction == NULL) {
+    g_err << "Can't get transaction pointer" << endl;
+    return NDBT_FAILED;
+  }
+  if(hugoTrans.setTransaction(pTransaction) != 0) {
+    g_err << "Set transaction error" << endl;
+    pNdb->closeTransaction(pTransaction);
+    return NDBT_FAILED;
+  }
+  if(hugoTrans.pkInsertRecord(pNdb, 0, num, 2) != 0) {
+    g_err << "pkInsertRecord error" << endl;
+    pNdb->closeTransaction(pTransaction);
+    return NDBT_FAILED;
+  }   
+  if(pTransaction->execute(Commit ) != 0) {
+    g_err << "Can't commit transaction delete" << endl;
+    return NDBT_FAILED;
+  }
+  hugoTrans.closeTransaction(pNdb);
+
+  // make sure backup have finish
+  NdbBackup backup;
+
+  // start log event
+  if(backup.startLogEvent() != 0) {
+    g_err << "Can't create log event" << endl;
+    return NDBT_FAILED;
+  }
+  NdbSleep_SecSleep(15);
+  int i = 0;
+  while (1) {
+    if (backup.checkBackupStatus() == 2) //complete
+      break;
+    else if (i == 15) {
+      g_err << "Backup timeout" << endl;
+      return NDBT_FAILED;
+    } else
+      NdbSleep_SecSleep(2);
+    i++;
+  }
+
+  return NDBT_OK;
+}
+
+int runVerifyUndoData(NDBT_Context* ctx, NDBT_Step* step){
+  int records = ctx->getNumRecords();
+  Ndb* pNdb = GETNDB(step);
+  int count = 0;
+  int num = 5;
+  if (records - 5 < 0)
+    num = 1;
+
+  const NdbDictionary::Table* tab = 
+    GETNDB(step)->getDictionary()->getTable(ctx->getTab()->getName());
+
+  if(tab == 0) {
+    g_err << " Can't find table" << endl;
+    return NDBT_FAILED;
+  }
+  
+  UtilTransactions utilTrans(* tab);
+  HugoTransactions hugoTrans(* tab);
+
+  // Check that there are as many records as we expected
+  if(utilTrans.selectCount(pNdb, 64, &count) != 0) {
+    g_err << "Can't get records count" << endl;
+    return NDBT_FAILED;
+  }
+
+  g_err << "count = " << count;
+  g_err << " records = " << records;
+  g_err << endl;
+
+  if (count != records) {
+    g_err << "The records count is not correct" << endl;
+    return NDBT_FAILED;
+  }
+
+  // make sure all the update data is there
+  NdbTransaction *pTransaction= pNdb->startTransaction();
+  if (pTransaction == NULL) {
+    g_err << "Can't get transaction pointer" << endl;
+    return NDBT_FAILED;
+  }
+  if(hugoTrans.setTransaction(pTransaction) != 0) {
+    g_err << "Set transaction error" << endl;
+    pNdb->closeTransaction(pTransaction);
+    return NDBT_FAILED;
+  }
+   if(hugoTrans.pkReadRecord(pNdb, 0, records, NdbOperation::LM_Read) != 0) {
+      g_err << "Can't read record" << endl;
+      return NDBT_FAILED;
+    }
+  if(hugoTrans.verifyUpdatesValue(0, records) != 0) {
+    g_err << "The records restored with undo log is not correct" << endl;
+    return NDBT_FAILED;
+  }
+  hugoTrans.closeTransaction(pNdb);
+
+  return NDBT_OK;
+}
+
+int
+runBug57650(NDBT_Context* ctx, NDBT_Step* step)
+{
+  NdbBackup backup;
+  NdbRestarter res;
+
+  int node0 = res.getNode(NdbRestarter::NS_RANDOM);
+  res.insertErrorInNode(node0, 5057);
+
+  unsigned backupId = 0;
+  if (backup.start(backupId) == -1)
+    return NDBT_FAILED;
+
+  res.insertErrorInAllNodes(5057);
+  int val2[] = { 7099 }; // Force LCP
+  res.dumpStateAllNodes(val2, 1);
+
+  NdbSleep_SecSleep(5);
+  res.waitClusterStarted();
+
+  res.insertErrorInAllNodes(0);
+
+  return NDBT_OK;
+}
 
 NDBT_TESTSUITE(testBackup);
 TESTCASE("BackupOne", 
@@ -550,7 +741,24 @@ TESTCASE("BackupBank",
   // TODO  STEP(runBankSum);
   STEP(runBackupBank);
   VERIFIER(runRestoreBankAndVerify);
-  //  FINALIZER(runDropBank);
+  FINALIZER(runDropBank);
+}
+TESTCASE("BackupUndoLog", 
+	 "Test for backup happen at start time\n"
+	 "1. Load table\n"
+	 "2. Start backup with wait started\n"
+	 "3. Insert, delete, update data during backup\n"
+	 "4. Drop tables and restart \n"
+	 "5. Restore\n"
+	 "6. Verify records of table\n"
+	 "7. Clear tables\n"){
+  INITIALIZER(runLoadTable);
+  INITIALIZER(runBackupUndoWaitStarted);
+  INITIALIZER(runChangeUndoDataDuringBackup);
+  INITIALIZER(runDropTablesRestart);
+  INITIALIZER(runRestoreOne);
+  VERIFIER(runVerifyUndoData);
+  FINALIZER(runClearTable);
 }
 TESTCASE("NFMaster", 
 	 "Test that backup behaves during node failiure\n"){
@@ -594,10 +802,15 @@ TESTCASE("FailSlave",
   STEP(runFail);
 
 }
+TESTCASE("Bug57650", "")
+{
+  INITIALIZER(runBug57650);
+}
 NDBT_TESTSUITE_END(testBackup);
 
 int main(int argc, const char** argv){
   ndb_init();
+  NDBT_TESTSUITE_INSTANCE(testBackup);
   return testBackup.execute(argc, argv);
 }
 
