@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,6 +27,7 @@
 #include <signaldata/DictTabInfo.hpp>
 #include <ndb_limits.h>
 #include <NdbAutoPtr.hpp>
+#include "../src/ndbapi/NdbDictionaryImpl.hpp"
 
 #include "../../../../sql/ha_ndbcluster_tables.h"
 extern NdbRecordPrintFormat g_ndbrecord_print_format;
@@ -117,7 +118,9 @@ public:
       break;
     default:
       abort();
+      break;
     }
+    return false; // Never reached
   }
 
   void twiddle_aligned(void* const data_ptr) const {
@@ -418,6 +421,37 @@ RestoreMetaData::readMetaTableDesc() {
 	   << dec << dst->getObjectId() << " " << dst->getPath() << endl;
     break;
   }
+  case DictTabInfo::HashMap:
+  {
+    NdbDictionary::HashMap * dst = new NdbDictionary::HashMap;
+    errcode =
+      NdbDictInterface::parseHashMapInfo(NdbHashMapImpl::getImpl(* dst),
+                                         (Uint32*)ptr, len);
+    if (errcode)
+      delete dst;
+    obj.m_objPtr = dst;
+
+    if (!m_hostByteOrder)
+    {
+      /**
+       * Bloddy byte-array, need to twiddle
+       */
+      Vector<Uint32> values;
+      Uint32 len = dst->getMapLen();
+      Uint32 zero = 0;
+      values.fill(len - 1, zero);
+      dst->getMapValues(values.getBase(), values.size());
+      for (Uint32 i = 0; i<len; i++)
+      {
+        values[i] = Twiddle16(values[i]);
+      }
+      dst->setMap(values.getBase(), values.size());
+    }
+
+    m_objects.push(obj, 0); // Put first
+    return true;
+    break;
+  }
   default:
     if (ga_skip_unknown_objects)
     {
@@ -472,6 +506,10 @@ end:
   return true;
 }
 
+#define OLD_NDB_REP_DB  "cluster"
+#define OLD_NDB_APPLY_TABLE "apply_status"
+#define OLD_NDB_SCHEMA_TABLE "schema"
+
 bool
 RestoreMetaData::markSysTables()
 {
@@ -485,6 +523,10 @@ RestoreMetaData::markSysTables()
         strcmp(tableName, "NDB$EVENTS_0") == 0 ||
         strcmp(tableName, "sys/def/SYSTAB_0") == 0 ||
         strcmp(tableName, "sys/def/NDB$EVENTS_0") == 0 ||
+        // index stats tables and indexes
+        strncmp(tableName, NDB_INDEX_STAT_PREFIX,
+                sizeof(NDB_INDEX_STAT_PREFIX)-1) == 0 ||
+        strstr(tableName, "/" NDB_INDEX_STAT_PREFIX) != 0 ||
         /*
           The following is for old MySQL versions,
            before we changed the database name of the tables from
@@ -782,8 +824,9 @@ RestoreDataIterator::free_bitfield_storage()
 Uint32
 RestoreDataIterator::get_free_bitfield_storage() const
 {
-  return (m_bitfield_storage_ptr + m_bitfield_storage_len) - 
-    m_bitfield_storage_curr_ptr;
+  
+  return Uint32((m_bitfield_storage_ptr + m_bitfield_storage_len) - 
+    m_bitfield_storage_curr_ptr);
 }
 
 Uint32*
@@ -1258,7 +1301,6 @@ BackupFile::BackupFile(void (* _free_data_callback)())
   : free_data_callback(_free_data_callback)
 {
   memset(&m_file,0,sizeof(m_file));
-  m_file.file = -1;
   m_path[0] = 0;
   m_fileName[0] = 0;
 
@@ -1269,48 +1311,44 @@ BackupFile::BackupFile(void (* _free_data_callback)())
 
   m_file_size = 0;
   m_file_pos = 0;
+  m_is_undolog = false;
 }
 
-BackupFile::~BackupFile(){
-  if(m_file.file > 1)
-  {
-    azclose(&m_file);
-    memset(&m_file,0,sizeof(m_file));
-  }
+BackupFile::~BackupFile()
+{
+  (void)ndbzclose(&m_file);
+
   if(m_buffer != 0)
     free(m_buffer);
 }
 
 bool
 BackupFile::openFile(){
-  if(m_file.file > 1){
-    azclose(&m_file);
-    m_file.file = 0;
-    m_file_size = 0;
-    m_file_pos = 0;
-  }
+  (void)ndbzclose(&m_file);
+  m_file_size = 0;
+  m_file_pos = 0;
 
   info.setLevel(254);
   info << "Opening file '" << m_fileName << "'\n";
-  int r= azopen(&m_file,m_fileName, O_RDONLY);
+  int r= ndbzopen(&m_file, m_fileName, O_RDONLY);
 
-  if (r==0)
+  if(r != 1)
+    return false;
+
+  size_t size;
+  if (ndbz_file_size(&m_file, &size) == 0)
   {
-    struct stat buf;
-    if (fstat(m_file.file, &buf) == 0)
-    {
-      m_file_size = (Uint64)buf.st_size;
-      info << "File size " << m_file_size << " bytes\n";
-    }
-    else
-    {
-      info << "Progress reporting degraded output since fstat failed,"
-           << "errno: " << errno << endl;
-      m_file_size = 0;
-    }
+    m_file_size = (Uint64)size;
+    info << "File size " << m_file_size << " bytes\n";
+  }
+  else
+  {
+    info << "Progress reporting degraded output since fstat failed,"
+         << "errno: " << errno << endl;
+    m_file_size = 0;
   }
 
-  return r != 0;
+  return true;
 }
 
 Uint32 BackupFile::buffer_get_ptr_ahead(void **p_buf_ptr, Uint32 size, Uint32 nmemb)
@@ -1322,21 +1360,107 @@ Uint32 BackupFile::buffer_get_ptr_ahead(void **p_buf_ptr, Uint32 size, Uint32 nm
       (*free_data_callback)();
 
     reset_buffers();
-    memcpy(m_buffer, m_buffer_ptr, m_buffer_data_left);
 
-    int error;
-    size_t r = azread(&m_file,
-                      ((char *)m_buffer) + m_buffer_data_left,
-                      m_buffer_sz - m_buffer_data_left, &error);
-    m_file_pos += r;
-    m_buffer_data_left += r;
-    m_buffer_ptr = m_buffer;
+    if (m_is_undolog)
+    {
+      /* move the left data to the end of buffer
+       */
+      size_t r = 0;
+      int error;
+      /* move the left data to the end of buffer
+       * m_buffer_ptr point the end of the left data. buffer_data_start point the start of left data
+       * m_buffer_data_left is the length of left data.
+       */
+      Uint64 file_left_entry_data = 0;
+      Uint32 buffer_free_space = m_buffer_sz - m_buffer_data_left;
+      void * buffer_end = (char *)m_buffer + m_buffer_sz;
+      void * buffer_data_start = (char *)m_buffer_ptr - m_buffer_data_left;
+
+      memmove((char *)buffer_end - m_buffer_data_left, buffer_data_start, m_buffer_data_left);
+      buffer_data_start = (char *)buffer_end - m_buffer_data_left;
+      /*
+       * For undo log file we should read log entris backwards from log file.
+       *   That mean the first entries should start at sizeof(m_fileHeader).
+       *   The end of the last entries should be the end of log file(EOF-1).
+       * If ther are entries left in log file to read.
+       *   m_file_pos should bigger than sizeof(m_fileHeader).
+       * If the length of left log entries less than the residual length of buffer,
+       *   we just need to read all the left entries from log file into the buffer.
+       *   and all the left entries in log file should been read into buffer. Or
+       * If the length of left entries is bigger than the residual length of buffer,
+       *   we should fill the buffer because the current buffer can't contain
+           all the left log entries, we should read more times.
+       * 
+       */
+      if (m_file_pos > sizeof(m_fileHeader))
+      {
+        /*
+         * We read(consume) data from the end of the buffer.
+         * If the left data is not enough for next read in buffer,
+         *   we move the residual data to the end of buffer.
+         *   Then we will fill the start of buffer with new data from log file.
+         * eg. If the buffer length is 10. "+" denotes useless content.
+         *                          top        end
+         *   Bytes in file        abcdefgh0123456789
+         *   Byte in buffer       0123456789             --after first read
+         *   Consume datas...     (6789) (2345)
+         *   Bytes in buffer      01++++++++             --after several consumes
+         *   Move data to end     ++++++++01
+         *   Bytes in buffer      abcdefgh01             --after second read
+         */
+	file_left_entry_data = m_file_pos - sizeof(m_fileHeader);
+        if (file_left_entry_data <= buffer_free_space)
+        {
+          /* All remaining data fits in space available in buffer. 
+	   * Read data into buffer before existing data.
+	   */
+          // Move to the start of data to be read
+          ndbzseek(&m_file, sizeof(m_fileHeader), SEEK_SET);
+          r = ndbzread(&m_file,
+                       (char *)buffer_data_start - file_left_entry_data,
+                       Uint32(file_left_entry_data),
+                       &error);
+          //move back
+          ndbzseek(&m_file, sizeof(m_fileHeader), SEEK_SET);
+        }
+        else
+        {
+	  // Fill remaing space at start of buffer with data from file.
+          ndbzseek(&m_file, m_file_pos-buffer_free_space, SEEK_SET);
+          r = ndbzread(&m_file, ((char *)m_buffer), buffer_free_space, &error);
+          ndbzseek(&m_file, m_file_pos-buffer_free_space, SEEK_SET);
+        }
+      }
+      m_file_pos -= r;
+      m_buffer_data_left += (Uint32)r;
+      //move to the end of buffer
+      m_buffer_ptr = buffer_end;
+    }
+    else
+    {
+      memmove(m_buffer, m_buffer_ptr, m_buffer_data_left);
+      int error;
+      Uint32 r = ndbzread(&m_file,
+                          ((char *)m_buffer) + m_buffer_data_left,
+                          m_buffer_sz - m_buffer_data_left, &error);
+      m_file_pos += r;
+      m_buffer_data_left += r;
+      m_buffer_ptr = m_buffer;
+    }
 
     if (sz > m_buffer_data_left)
       sz = size * (m_buffer_data_left / size);
   }
 
-  *p_buf_ptr = m_buffer_ptr;
+  /*
+   * For undolog, the m_buffer_ptr points to the end of the left data.
+   * After we get data from the end of buffer, the data-end move forward.
+   *   So we should move m_buffer_ptr to the right place.
+   */
+  if(m_is_undolog)
+    *p_buf_ptr = (char *)m_buffer_ptr - sz;
+  else
+    *p_buf_ptr = m_buffer_ptr;
 
   return sz/size;
 }
@@ -1344,8 +1468,19 @@ Uint32 BackupFile::buffer_get_ptr(void **p_buf_ptr, Uint32 size, Uint32 nmemb)
 {
   Uint32 r = buffer_get_ptr_ahead(p_buf_ptr, size, nmemb);
 
-  m_buffer_ptr = ((char*)m_buffer_ptr)+(r*size);
-  m_buffer_data_left -= (r*size);
+  if(m_is_undolog)
+  {
+    /* we read from end of buffer to start of buffer.
+     * m_buffer_ptr keep at the end of real data in buffer.
+     */
+    m_buffer_ptr = ((char*)m_buffer_ptr)-(r*size);
+    m_buffer_data_left -= (r*size);
+  }
+  else
+  {
+    m_buffer_ptr = ((char*)m_buffer_ptr)+(r*size);
+    m_buffer_data_left -= (r*size);
+  }
 
   return r;
 }
@@ -1407,10 +1542,10 @@ void
 BackupFile::setName(const char * p, const char * n){
   const Uint32 sz = sizeof(m_path);
   if(p != 0 && strlen(p) > 0){
-    if(p[strlen(p)-1] == '/'){
+    if(p[strlen(p)-1] == DIR_SEPARATOR[0]){
       BaseString::snprintf(m_path, sz, "%s", p);
     } else {
-      BaseString::snprintf(m_path, sz, "%s%s", p, "/");
+      BaseString::snprintf(m_path, sz, "%s%s", p, DIR_SEPARATOR);
     }
   } else {
     m_path[0] = 0;
@@ -1479,10 +1614,30 @@ BackupFile::readHeader(){
   debug << "ByteOrder is " << m_fileHeader.ByteOrder << endl;
   debug << "magicByteOrder is " << magicByteOrder << endl;
   
-  if (m_fileHeader.FileType != m_expectedFileHeader.FileType){
+
+  if (m_fileHeader.FileType != m_expectedFileHeader.FileType &&
+      !(m_expectedFileHeader.FileType == BackupFormat::LOG_FILE &&
+      m_fileHeader.FileType == BackupFormat::UNDO_FILE)){
+    // UNDO_FILE will do in case where we expect LOG_FILE
     abort();
   }
   
+  if(m_fileHeader.FileType == BackupFormat::UNDO_FILE){
+      m_is_undolog = true;
+      /* move pointer to end of data part. 
+         move 4 bytes from the end of file 
+         because footer contain 4 bytes 0 at the end of file.
+         we discard the remain data stored in m_buffer.
+      */
+      size_t size;
+      if (ndbz_file_size(&m_file, &size) == 0)
+        m_file_size = (Uint64)size;
+      ndbzseek(&m_file, 4, SEEK_END);
+      m_file_pos = m_file_size - 4;
+      m_buffer_data_left = 0;
+      m_buffer_ptr = m_buffer;
+  }
+
   // Check for BackupFormat::FileHeader::ByteOrder if swapping is needed
   if (m_fileHeader.ByteOrder == magicByteOrder) {
     m_hostByteOrder = true;
@@ -1734,9 +1889,24 @@ RestoreLogIterator::getNextLogEntry(int & res) {
   do {
     Uint32 len;
     Uint32 *logEntryPtr;
-    if (buffer_read_ahead(&len, sizeof(Uint32), 1) != 1){
-      res= -1;
-      return 0;
+    if(m_is_undolog){
+      int read_result = 0;
+      read_result = buffer_read(&len, sizeof(Uint32), 1);
+      //no more log data to read
+      if (read_result == 0 ) {
+        res = 0;
+        return 0;
+      }
+      if (read_result != 1) {
+        res= -1;
+        return 0;
+      }
+    }
+    else{
+      if (buffer_read_ahead(&len, sizeof(Uint32), 1) != 1){
+        res= -1;
+        return 0;
+      }
     }
     len= ntohl(len);
 
@@ -1791,15 +1961,27 @@ RestoreLogIterator::getNextLogEntry(int & res) {
   } while(m_last_gci > stopGCP + 1);
 
   m_logEntry.m_table = m_metaData.getTable(tableId);
+  /* We should 'invert' the operation type when we restore an Undo log.
+   *   To undo an insert operation, a delete is required.
+   *   To undo a delete operation, an insert is required.
+   * The backup have collected 'before values' for undoing 'delete+update' to make this work.
+   * To undo insert, we only need primary key.
+   */
   switch(triggerEvent){
   case TriggerEvent::TE_INSERT:
-    m_logEntry.m_type = LogEntry::LE_INSERT;
+    if(m_is_undolog)
+      m_logEntry.m_type = LogEntry::LE_DELETE;
+    else
+      m_logEntry.m_type = LogEntry::LE_INSERT;
     break;
   case TriggerEvent::TE_UPDATE:
     m_logEntry.m_type = LogEntry::LE_UPDATE;
     break;
   case TriggerEvent::TE_DELETE:
-    m_logEntry.m_type = LogEntry::LE_DELETE;
+    if(m_is_undolog)
+      m_logEntry.m_type = LogEntry::LE_INSERT;
+    else
+      m_logEntry.m_type = LogEntry::LE_DELETE;
     break;
   default:
     res = -1;
